import numpy as np
import warnings
from traitlets import Any, Bool, List, Unicode, observe
from astropy import units as u
from astropy.io import fits
from astropy.nddata import NDDataArray

try:
    from stdatamodels.jwst.datamodels import Level1bModel
except ImportError:
    Level1bModel = None

try:
    from roman_datamodels.datamodels import RampModel, ScienceRawModel
except ImportError:
    RampModel = None
    ScienceRawModel = None

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import loader_importer_registry, viewer_registry
from jdaviz.core.loaders.importers import BaseImporterToDataCollection
from jdaviz.core.template_mixin import (AutoTextField,
                                        SelectPluginComponent,
                                        ViewerSelectCreateNew)
from jdaviz.core.user_api import ImporterUserApi
from jdaviz.utils import (standardize_metadata,
                          standardize_roman_metadata,
                          PRIHDR_KEY)


__all__ = ['RampImporter']


def move_group_axis_last(x):
    # swap axes per the conventions of ramp cubes
    # (group axis comes first) and the default in
    # rampviz (group axis expected last)
    return np.transpose(x, (1, 2, 0))


@loader_importer_registry('Ramp')
class RampImporter(BaseImporterToDataCollection):
    template_file = __file__, "./ramp.vue"

    # INTEGRATION SELECT
    integration_items = List().tag(sync=True)
    integration_selected = Unicode().tag(sync=True)

    # DIFF CUBE
    diff_data_label_value = Unicode().tag(sync=True)
    diff_data_label_default = Unicode().tag(sync=True)
    diff_data_label_auto = Bool(True).tag(sync=True)
    diff_data_label_invalid_msg = Unicode().tag(sync=True)

    # DIFF VIEWER
    diff_viewer_create_new_items = List([]).tag(sync=True)
    diff_viewer_create_new_selected = Unicode().tag(sync=True)
    diff_viewer_items = List([]).tag(sync=True)
    diff_viewer_selected = Any([]).tag(sync=True)
    diff_viewer_multiselect = Bool(True).tag(sync=True)

    diff_viewer_label_value = Unicode().tag(sync=True)
    diff_viewer_label_default = Unicode().tag(sync=True)
    diff_viewer_label_auto = Bool(True).tag(sync=True)
    diff_viewer_label_invalid_msg = Unicode().tag(sync=True)

    # EXTRACTED INTEGRATION
    auto_extract = Bool(True).tag(sync=True)
    function_items = List().tag(sync=True)
    function_selected = Unicode('Sum').tag(sync=True)

    ext_data_label_value = Unicode().tag(sync=True)
    ext_data_label_default = Unicode().tag(sync=True)
    ext_data_label_auto = Bool(True).tag(sync=True)
    ext_data_label_invalid_msg = Unicode().tag(sync=True)

    # INTEGRATION VIEWER
    ext_viewer_create_new_items = List([]).tag(sync=True)
    ext_viewer_create_new_selected = Unicode().tag(sync=True)
    ext_viewer_items = List([]).tag(sync=True)
    ext_viewer_selected = Any([]).tag(sync=True)
    ext_viewer_multiselect = Bool(True).tag(sync=True)

    ext_viewer_label_value = Unicode().tag(sync=True)
    ext_viewer_label_default = Unicode().tag(sync=True)
    ext_viewer_label_auto = Bool(True).tag(sync=True)
    ext_viewer_label_invalid_msg = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def viewer_in_registry_names(supported_viewers):
            def viewer_filter(viewer):
                classes = [viewer_registry.members.get(item.get('reference')).get('cls')
                           for item in supported_viewers]
                return isinstance(viewer, tuple(classes))
            return viewer_filter

        if Level1bModel is not None and isinstance(self.input, Level1bModel):
            integration_options = [str(i) for i in range(len(self.input.data))]
        elif isinstance(self.input, fits.HDUList):
            # TODO: this will need to be adjusted if adding extension selection
            integration_options = [str(i) for i in range(len(self.input[1].data))]
        else:
            integration_options = []
        self.integration = SelectPluginComponent(self,
                                                 items='integration_items',
                                                 selected='integration_selected',
                                                 manual_options=integration_options)

        # RAMP GROUP CUBE
        if self.default_data_label_from_resolver:
            self.data_label_default = self.default_data_label_from_resolver + '[DATA]'
        elif self.config == 'rampviz':
            # TODO: backwards compaibility for other inputs
            self.data_label_default = self.input.__class__.__name__ + '[DATA]'
        else:
            self.data_label_default = 'Ramp[DATA]'

        if self.config == 'rampviz':
            self.viewer.selected = ['group-viewer']

        # DIFF CUBE
        self.diff_data_label = AutoTextField(self,
                                             'diff_data_label_value',
                                             'diff_data_label_default',
                                             'diff_data_label_auto',
                                             'diff_data_label_invalid_msg')

        self.diff_viewer = ViewerSelectCreateNew(self,
                                                 'diff_viewer_items',
                                                 'diff_viewer_selected',
                                                 'diff_viewer_create_new_items',
                                                 'diff_viewer_create_new_selected',
                                                 'diff_viewer_label_value',
                                                 'diff_viewer_label_default',
                                                 'diff_viewer_label_auto',
                                                 'diff_viewer_label_invalid_msg',
                                                 multiselect='diff_viewer_multiselect',
                                                 default_mode='empty')
        supported_viewers = [{'label': '3D Ramp Diff',
                              'reference': 'rampviz-image-viewer'}]
        if self.app.config == 'deconfigged':
            self.diff_viewer_create_new_items = supported_viewers
        self.diff_viewer.add_filter(viewer_in_registry_names(supported_viewers))
        if self.config == 'rampviz':
            self.diff_viewer.selected = ['diff-viewer']
        else:
            self.diff_viewer.select_default()

        # RAMP INTEGRATION EXTRACTION
        self.function = SelectPluginComponent(
            self,
            items='function_items',
            selected='function_selected',
            manual_options=['Mean', 'Median', 'Min', 'Max', 'Sum']
        )
        # the default collapse function in the profile viewer is "sum",
        # but for ramp files, "median" is more useful:
        self.function.selected = 'Median'

        self.ext_data_label = AutoTextField(self,
                                            'ext_data_label_value',
                                            'ext_data_label_default',
                                            'ext_data_label_auto',
                                            'ext_data_label_invalid_msg')

        self.ext_viewer = ViewerSelectCreateNew(self,
                                                'ext_viewer_items',
                                                'ext_viewer_selected',
                                                'ext_viewer_create_new_items',
                                                'ext_viewer_create_new_selected',
                                                'ext_viewer_label_value',
                                                'ext_viewer_label_default',
                                                'ext_viewer_label_auto',
                                                'ext_viewer_label_invalid_msg',
                                                multiselect='ext_viewer_multiselect',  # noqa
                                                default_mode='empty')
        supported_viewers = [{'label': 'Ramp Integration',
                              'reference': 'rampviz-profile-viewer'}]
        if self.app.config == 'deconfigged':
            self.ext_viewer_create_new_items = supported_viewers
        self.ext_viewer.add_filter(viewer_in_registry_names(supported_viewers))
        if self.config == 'rampviz':
            self.ext_viewer.selected = ['integration-viewer']
        else:
            self.ext_viewer.select_default()

    @staticmethod
    def _get_supported_viewers():
        return [{'label': '3D Ramp', 'reference': 'rampviz-image-viewer'}]

    @property
    def user_api(self):
        expose = ['diff_data_label', 'diff_viewer',
                  'auto_extract', 'function', 'ext_data_label', 'ext_viewer']
        if len(self.integration_items) > 0:
            expose += ['integration']
        return ImporterUserApi(self, expose)

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'rampviz'):
            # NOTE: temporary during deconfig process
            return False

        # Filter out None types from isinstance check (optional dependencies)
        valid_types = tuple(
            t for t in (Level1bModel,
                        RampModel, ScienceRawModel,
                        fits.HDUList,
                        np.ndarray)
            if t is not None
        )
        if not isinstance(self.input, valid_types):
            return False

        if isinstance(self.input, fits.HDUList) and self.input[1].header['NAXIS'] != 4:
            return False

        try:
            self.output
        except Exception:
            return False
        return True

    @observe('data_label_value', 'function_selected')
    def _data_label_changed(self, msg={}):
        base = self.data_label_value.strip('[DATA]').strip()
        self.diff_data_label_default = f"{base}[DIFF]"
        self.ext_data_label_default = f"{base} ({self.function_selected.lower()})"

    @property
    def output(self):
        integration = 0  # TODO: integration/extension select

        # NOTE: each if-statement should provide meta and ramp_data
        # if there is specific handling for flux_unit, ramp_data should
        # be a quantity with the unit attached
        if Level1bModel is not None and isinstance(self.input, Level1bModel):
            meta = standardize_metadata({
                key: value for key, value in self.input.to_flat_dict(
                    include_arrays=False)
                .items()
                if key.startswith('meta')
            })

            ramp_data = self.input.data[integration]
        elif (RampModel is not None and ScienceRawModel is not None
              and isinstance(self.input, (RampModel, ScienceRawModel))):
            meta = standardize_roman_metadata(self.input)
            ramp_data = self.input.data
        elif isinstance(self.input, fits.HDUList):
            # TODO: extension selection (didn't exist previously)
            hdulist = self.input
            hdu = hdulist[1]  # extension containing the ramp

            meta = standardize_metadata(hdu.header)
            if hdu.name != 'PRIMARY' and 'PRIMARY' in hdulist:
                meta[PRIHDR_KEY] = standardize_metadata(hdulist['PRIMARY'].header)

            if 'BUNIT' in hdu.header:
                try:
                    flux_unit = u.Unit(hdu.header['BUNIT'])
                except Exception:
                    warnings.warn("Invalid BUNIT, using DN as data unit", UserWarning)
                    flux_unit = u.DN
            else:
                warnings.warn("Invalid BUNIT, using DN as data unit", UserWarning)
                flux_unit = u.DN

            # index the ramp array by the integration to load. returns all groups and pixels.
            # cast from uint16 to integers:
            ramp_data = hdu.data[integration].astype(int) * flux_unit
        elif isinstance(self.input, np.ndarray):
            meta = {}
            ramp_data = self.input
        else:
            raise NotImplementedError(f"Unsupported input for RampImporter: {type(self.input)}")

        # last axis is the group axis, first two are spatial axes:
        diff_data = np.vstack([
            # begin with a group of zeros, so
            # that `diff_data.ndim == data.ndim`
            np.zeros((1, *ramp_data[0].shape)),
            np.diff(ramp_data, axis=0)
        ])

        # if the ramp cube has no units, assume DN:
        flux_unit = getattr(ramp_data, 'unit', u.DN)

        ramp_cube = NDDataArray(move_group_axis_last(ramp_data),
                                unit=flux_unit,
                                meta=meta)
        diff_cube = NDDataArray(move_group_axis_last(diff_data),
                                unit=flux_unit,
                                meta=meta)

        return ramp_cube, diff_cube

    def __call__(self):
        # get a copy of all requested data-labels before additional data entries changes defaults
        data_label = self.data_label_value
        diff_data_label = self.diff_data_label_value
        ext_data_label = self.ext_data_label_value

        ramp_cube, diff_cube = self.output

        ramp_cube.meta['_ramp_type'] = 'group'
        self.add_to_data_collection(ramp_cube,
                                    data_label,
                                    viewer_select=self.viewer)
        # TODO: this will need to be removed when removing restriction of a single flux cube
        self.app._jdaviz_helper._loaded_flux_cube = self.app.data_collection[data_label]
        if not hasattr(self.app._jdaviz_helper, 'cube_cache'):
            self.app._jdaviz_helper.cube_cache = {}
        self.app._jdaviz_helper.cube_cache[data_label] = ramp_cube

        diff_cube.meta['_ramp_type'] = 'diff'
        self.add_to_data_collection(diff_cube,
                                    diff_data_label,
                                    viewer_select=self.diff_viewer)
        self.app._jdaviz_helper.cube_cache[diff_data_label] = diff_cube

        if not self.auto_extract:
            return

        try:
            rext = self.app.get_tray_item_from_name('ramp-extraction')
            ext = rext._extract_in_new_instance(dataset=data_label,
                                                function=self.function.selected,
                                                auto_update=False,
                                                add_data=False)
            # we'll add the data manually instead of through add_results_from_plugin
            # but still want to preserve the plugin metadata
            ext.meta['plugin'] = rext._plugin_name
        except Exception:
            ext = None
            msg = SnackbarMessage(
                "Automatic ramp extraction failed. See the ramp extraction"
                " plugin to perform a custom extraction",
                color='error', sender=self, timeout=10000)
        else:
            msg = SnackbarMessage(
                "The extracted 1D ramp integration was generated automatically."
                " See the ramp extraction plugin for details or to"
                " perform a custom extraction.",
                color='warning', sender=self, timeout=10000)
        self.app.hub.broadcast(msg)

        if ext is not None:
            ext_viewer_selected = self.ext_viewer.create_new.selected if self.ext_viewer.create_new.selected != '' else self.ext_viewer.selected  # noqa
            self.app._jdaviz_helper.load(ext, format='Ramp Integration',
                                         data_label=ext_data_label,
                                         viewer=ext_viewer_selected)



from functools import cached_property
from astropy.io import fits
from astropy.io.fits import VerifyError
from astropy.table import Table, QTable
import numpy as np

from jdaviz.core.loaders.parsers import BaseParser
from jdaviz.core.registries import loader_parser_registry


__all__ = ['AstropyTableParser']


@loader_parser_registry('astropy.Table')
class AstropyTableParser(BaseParser):

    @property
    def is_valid(self):

        if self.app.config not in ('deconfigged', 'imviz', 'mastviz'):
            # NOTE: temporary during deconfig process
            return False

        if self.input is None:
            return False

        # first, check if this is a Table or QTable object
        if isinstance(self.input, (Table, QTable)):
            if len(self.input) == 0:
                return False
            else:
                return True

        # fits files can be sucessfully opened with table.read
        # try to reject fits files from being validated as catalogs
        # by trying to open the input with fits.open, and rejecting
        # it as a catalog type if it opens sucessfully
        # eventually we may want to accept BinTableHDU/TableHDU
        # inside fits so this logic should be improved then
        if isinstance(self.input, (fits.ImageHDU, fits.HDUList, fits.PrimaryHDU, fits.CompImageHDU)):  # noqa
            return False
        elif isinstance(self.input, np.ndarray):
            # arrays can be loaded as tables, skip these so images/spectra
            # aren't mis-identified as catalogs
            return False
        try:
            f = fits.open(self.input)
            f.close()
            return False
        except (OSError, FileNotFoundError, VerifyError):  # noqa
            # if we can't open as fits, continue checking if catalog
            pass

        # next, see if this is a catalog written to a file
        try:
            table = self.output
        except Exception:
            return False
        else:
            return len(table) > 0

    @cached_property
    def output(self):
        return QTable.read(self.input, format='ascii')



from astropy.coordinates import SkyCoord
from astropy import units as u


from traitlets import Unicode, List

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import loader_resolver_registry
from jdaviz.core.template_mixin import (
    SelectPluginComponent,
    with_spinner,
)
from jdaviz.core.loaders.resolvers import BaseConeSearchResolver
from jdaviz.core.user_api import LoaderUserApi

__all__ = ["AstroqueryResolver"]


@loader_resolver_registry("astroquery")
class AstroqueryResolver(BaseConeSearchResolver):
    template_file = __file__, "astroquery.vue"

    telescope_items = List([]).tag(sync=True)
    telescope_selected = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.telescope = SelectPluginComponent(
            self, items="telescope_items", selected="telescope_selected",
            manual_options=['SDSS', 'Gaia']
        )

    @property
    def user_api(self):
        return LoaderUserApi(
            self,
            expose=[
                "viewer", "coordframe", "radius", "radius_unit",
                "source",
                "telescope",
                "max_results",
                "query_archive"
            ],
        )

    @with_spinner(spinner_traitlet="results_loading")
    def query_archive(self):
        skycoord_center = SkyCoord.from_name(self.source, frame=self.coordframe.selected)
        radius = self.radius * u.Unit(self.radius_unit.selected)

        if self.telescope.selected == 'SDSS':
            from astroquery.sdss import SDSS

            r_max = 3 * u.arcmin
            if radius > r_max:  # SDSS now has radius max limit
                self.hub.broadcast(SnackbarMessage(
                    f"Radius for {self.telescope.selected} has max radius of {r_max}\' but got "
                    f"{radius.to(u.arcmin)}, using {r_max}.",
                    color='warning', sender=self))
                radius = r_max

            # queries the region (based on the provided center point and radius)
            # finds all the sources in that region
            try:
                output = SDSS.query_region(skycoord_center, radius=radius,
                                           data_release=17)
            except Exception as e:  # nosec
                errmsg = (f"Failed to query {self.telescope.selected} with c={skycoord_center} and "
                          f"r={radius}: {repr(e)}")
                self.hub.broadcast(SnackbarMessage(errmsg, color='error',
                                                   sender=self,
                                                   traceback=e))
                output = None  # will force returned_max_results = False, returned_no_results = True
        elif self.telescope.selected == 'Gaia':
            from astroquery.gaia import Gaia

            Gaia.ROW_LIMIT = self.max_results
            output = Gaia.query_object(skycoord_center, radius=radius)
        else:
            raise NotImplementedError(f"Querying for {self.telescope.selected} is not supported.")

        if output is not None and len(output) > self.max_results:
            output = output[:self.max_results]
            self.returned_max_results = True
        else:
            self.returned_max_results = False
        if output is None or len(output) == 0:
            self.returned_no_results = True
        else:
            self.returned_no_results = False
        self._output = output

        self._resolver_input_updated()

    def vue_query_archive(self, _=None):
        self.query_archive()

    def parse_input(self):
        return self._output



from jdaviz.configs.imviz.tests.utils import BaseImviz_WCS_WCS
import pytest
import re


# This applies to all viz but testing with Imviz should be enough.
class TestImviz_WCS_WCS(BaseImviz_WCS_WCS):
    def test_imviz_zoom_level(self):
        v = self.imviz.viewers['imviz-0']
        assert v._obj.glue_viewer.state.x_min == -0.5
        assert v._obj.glue_viewer.state.x_max == 9.5

        v.zoom(2)

        assert v._obj.glue_viewer.state.x_min == 1.5
        assert v._obj.glue_viewer.state.x_max == 6.5

    def test_imviz_viewers(self):
        self.imviz.create_image_viewer()
        self.imviz.create_image_viewer()

        # regression test for https://github.com/spacetelescope/jdaviz/pull/2624
        assert len(self.imviz.viewers) == 3


def test_specviz_zoom_level(specviz_helper):
    v = specviz_helper.viewers['spectrum-viewer']
    v.set_limits(x_min=1, x_max=2, y_min=1, y_max=2)
    assert v._obj.glue_viewer.state.x_min == 1
    assert v._obj.glue_viewer.state.x_max == 2
    assert v._obj.glue_viewer.state.y_min == 1
    assert v._obj.glue_viewer.state.y_max == 2


def test_specviz_data_labels(specviz_helper, spectrum1d):
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    assert specviz_helper.data_labels == [label]
    assert specviz_helper.viewers['spectrum-viewer'].data_menu.data_labels_loaded == [label]
    assert specviz_helper.viewers['spectrum-viewer'].data_menu.data_labels_visible == [label]


def test_toggle_api_hints(specviz_helper):
    assert specviz_helper.app.state.show_api_hints is False
    specviz_helper.toggle_api_hints()
    assert specviz_helper.app.state.show_api_hints is True
    specviz_helper.toggle_api_hints(True)
    assert specviz_helper.app.state.show_api_hints is True
    specviz_helper.toggle_api_hints()
    assert specviz_helper.app.state.show_api_hints is False


def test_wildcard_match_sources(specviz_helper, premade_spectrum_list):
    """
    Test wildcard matching for source selection in Specviz. This tests setting
    the selection directly as opposed to using ``load``, via ``ldr.importer.sources``
    (whereas in the following test this is done through ``user_api.extension``, same idea).
    """
    default_choices = ['1D Spectrum at index: 0',
                       '1D Spectrum at index: 1',
                       'Exposure 0, Source ID: 0000',
                       'Exposure 0, Source ID: 1111',
                       'Exposure 1, Source ID: 1111']

    # Testing directly
    ldr = specviz_helper.loaders['object']
    ldr.object = premade_spectrum_list
    selection_obj = ldr.importer.sources
    assert selection_obj.selected == [default_choices[0]]
    assert selection_obj.choices == default_choices
    # Resetting to empty
    selection_obj.selected = []

    err_str1 = "not all items in"
    err_str2 = f"are one of {selection_obj.choices}, reverting selection to []"
    with pytest.raises(ValueError,
                       match=re.escape(f"{err_str1} ['bad *'] {err_str2}")):
        ldr.importer.sources = 'bad *'

    with pytest.raises(ValueError,
                       match=re.escape(f"{err_str1} ['bad *', '* result'] {err_str2}")):
        ldr.importer.sources = ['bad *', '* result']

    with pytest.raises(ValueError,
                       match=re.escape(f"{err_str1} ['another', 'bad * result'] {err_str2}")):
        ldr.importer.sources = ['another', 'bad * result']

    # Check that selected is still/reverted successfully to []
    assert selection_obj.selected == []

    # This should get set to True automatically when multiple selections are made
    selection_obj.multiselect = False
    ldr.importer._obj.user_api.sources = '*'
    assert selection_obj.selected == selection_obj.choices
    assert selection_obj.multiselect is True


def test_wildcard_match_extension(imviz_helper, multi_extension_image_hdu_wcs):
    """
    Test wildcard matching for source selection in Specviz. This tests setting
    the selection directly as opposed to using ``load``, via
    ``ldr.importer._obj.user_api.extensions`` (whereas in the previous test this is
    done through ``user_api.sources``, same idea).
    """
    default_choices = ['1: [SCI,1]',
                       '2: [MASK,1]',
                       '3: [ERR,1]',
                       '4: [DQ,1]']

    # Testing directly
    ldr = imviz_helper.loaders['object']
    ldr.object = multi_extension_image_hdu_wcs
    selection_obj = ldr.importer.extension

    # Default selection
    assert selection_obj.selected == [default_choices[0]]

    # Resetting to []
    # Note this can't be done by setting selected = [], is this intentional?
    selection_obj.selected.pop(0)
    assert selection_obj.selected == []
    assert selection_obj.choices == default_choices

    err_str1 = "not all items in"
    err_str2 = f"are one of {selection_obj.choices}, reverting selection to []"
    with pytest.raises(ValueError,
                       match=re.escape(f"{err_str1} ['bad *'] {err_str2}")):
        ldr.importer._obj.user_api.extension = 'bad *'

    with pytest.raises(ValueError,
                       match=re.escape(f"{err_str1} ['bad *', '* result'] {err_str2}")):
        ldr.importer._obj.user_api.extension = ['bad *', '* result']

    with pytest.raises(ValueError,
                       match=re.escape(f"{err_str1} ['another', 'bad * result'] {err_str2}")):
        ldr.importer._obj.user_api.extension = ['another', 'bad * result']

    # Check that selected is still/reverted successfully to []
    assert selection_obj.selected == []

    # This should get set to True automatically when multiple selections are made
    selection_obj.multiselect = False
    ldr.importer._obj.user_api.extension = '*'
    assert selection_obj.selected == selection_obj.choices
    assert selection_obj.multiselect is True


def test_viewer_create_new(deconfigged_helper, spectrum1d):
    assert len(deconfigged_helper.new_viewers.keys()) == 0
    # passing [] should not load into a new viewer nor should it create a new viewer
    deconfigged_helper.load(spectrum1d, format='1D Spectrum', viewer=[], data_label='data1')
    assert len(deconfigged_helper.app.data_collection) == 1
    assert len(deconfigged_helper.viewers) == 0
    assert len(deconfigged_helper.new_viewers.keys()) > 0

    # passing nothing when there are no viewers should create a new viewer
    deconfigged_helper.load(spectrum1d, format='1D Spectrum', data_label='data2')
    assert len(deconfigged_helper.app.data_collection) == 2
    assert len(deconfigged_helper.viewers) == 1
    assert len(deconfigged_helper.viewers['1D Spectrum'].data_menu.layer.choices) == 1

    # passing nothing when there is a viewer should default to loading into that viewer
    deconfigged_helper.load(spectrum1d, format='1D Spectrum', data_label='data3')
    assert len(deconfigged_helper.app.data_collection) == 3
    assert len(deconfigged_helper.viewers) == 1
    assert len(deconfigged_helper.viewers['1D Spectrum'].data_menu.layer.choices) == 2

    # passing a string of a viewer that does not exist should create a viewer with that label
    deconfigged_helper.load(spectrum1d, format='1D Spectrum', viewer='user-defined-viewer', data_label='data4')  # noqa
    assert len(deconfigged_helper.app.data_collection) == 4
    assert len(deconfigged_helper.viewers) == 2
    assert len(deconfigged_helper.viewers['1D Spectrum'].data_menu.layer.choices) == 2
    assert len(deconfigged_helper.viewers['user-defined-viewer'].data_menu.layer.choices) == 1


@pytest.mark.parametrize(
    ("selection", "matches"), [
        ('*', (0, 1, 2, 3)),
        (('*', '*:*'), (0, 1, 2, 3)),
        ('1:*', (0,)),
        ('*S*', (0, 1)),
        (('*ERR*', '*DQ*'), (2, 3)),
        # Brackets should be sanitized, if not this will fail
        ('?: [SCI,1]', (0,)),
        ('?:*', (0, 1, 2, 3)),
    ])
def test_wildcard_match_through_load(imviz_helper, multi_extension_image_hdu_wcs,
                                     selection, matches):
    data_labels = ['Image[SCI,1]',
                   'Image[MASK,1]',
                   'Image[ERR,1]',
                   'Image[DQ,1]']

    # Through load
    imviz_helper.load(multi_extension_image_hdu_wcs, extension=selection)
    assert imviz_helper.data_labels == [data_labels[i] for i in matches]



import warnings

import numpy as np
from astropy import units as u
from specutils import Spectrum

from jdaviz.core.custom_units_and_equivs import PIX2, _eqv_flux_to_sb_pixel
from jdaviz.core.unit_conversion_utils import check_if_unit_is_per_solid_angle

__all__ = []


def _return_spectrum_with_correct_units(flux, wcs, metadata, data_type=None,
                                        target_wave_unit=None, hdulist=None,
                                        uncertainty=None, mask=None, apply_pix2=False,
                                        spectral_axis=None):
    """Upstream issue of WCS not using the correct units for data must be fixed here.
    Issue: https://github.com/astropy/astropy/issues/3658.

    Also converts flux units to flux/pix2 solid angle units, if `flux` is not a surface
    brightness and `apply_pix2` is True.
    """
    # handle scale factors when they are included in the unit
    # (has to be done before Spectrum creation)
    if not np.isclose(flux.unit.scale, 1, rtol=1e-5):
        flux = flux.to(flux.unit / flux.unit.scale)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            'ignore', message='Input WCS indicates that the spectral axis is not last',
            category=UserWarning)
        if spectral_axis is None:
            sc = Spectrum(flux=flux, wcs=wcs, uncertainty=uncertainty, mask=mask, meta=metadata)
        else:
            sc = Spectrum(flux=flux, spectral_axis=spectral_axis,
                          uncertainty=uncertainty, mask=mask, meta=metadata)

    # convert flux and uncertainty to per-pix2 if input is not a surface brightness
    target_flux_unit = None
    if (apply_pix2 and (data_type != "mask") and
            (not check_if_unit_is_per_solid_angle(flux.unit))):
        target_flux_unit = flux.unit / PIX2
    elif check_if_unit_is_per_solid_angle(flux.unit, return_unit=True) == "spaxel":
        # We need to convert spaxel to pixel squared, since spaxel isn't fully supported by astropy
        # This is horribly ugly but just multiplying by u.Unit("spaxel") doesn't work
        target_flux_unit = flux.unit * u.Unit('spaxel') / PIX2

    if target_wave_unit is None and hdulist is not None:
        found_target = False
        for ext in ('SCI', 'FLUX', 'PRIMARY', 'DATA'):  # In priority order
            if found_target:
                break
            if ext not in hdulist:
                continue
            hdr = hdulist[ext].header
            # The WCS could be swapped or unswapped.
            for cunit_num in (3, 1):
                cunit_key = f"CUNIT{cunit_num}"
                ctype_key = f"CTYPE{cunit_num}"
                if cunit_key in hdr and 'WAV' in hdr[ctype_key]:
                    target_wave_unit = u.Unit(hdr[cunit_key])
                    found_target = True
                    break

    if target_wave_unit == sc.spectral_axis.unit:
        target_wave_unit = None

    if (target_wave_unit is None) and (target_flux_unit is None):  # Nothing to convert
        new_sc = sc
    elif target_flux_unit is None:  # Convert wavelength only
        new_sc = sc.with_spectral_axis_unit(target_wave_unit)
    elif target_wave_unit is None:  # Convert flux only and only PIX2 stuff
        new_sc = sc.with_flux_unit(target_flux_unit, equivalencies=_eqv_flux_to_sb_pixel())
    else:  # Convert both
        new_sc = sc.with_spectral_axis_and_flux_units(
            target_wave_unit, target_flux_unit, flux_equivalencies=_eqv_flux_to_sb_pixel())

    if target_wave_unit is not None:
        new_sc.meta['_orig_spec'] = sc  # Need this for later

    return new_sc



import ipyvuetify as v
from echo import CallbackProperty

from glue.viewers.image.layer_artist import ImageLayerArtist
from glue.viewers.matplotlib.state import DeferredDrawCallbackProperty as DDCProperty
from glue.viewers.image.state import ImageLayerState


__all__ = ['SonifiedLayerStateWidget', 'SonifiedDataLayerArtist']


class SonifiedLayerState(ImageLayerState):
    volume = DDCProperty(100, docstring='The volume level of this layer')
    audible = CallbackProperty(True, 'whether audio will be allowed to play')

    def __init__(self, *args, **kwargs):

        super(SonifiedLayerState, self).__init__(*args, **kwargs)
        self.alpha = 0
        self.previous_volume = self.volume

    def _update_volume(self, ignore=None):
        if not self.audible:
            self.previous_volume = self.volume
            self.volume = 0
        else:
            self.volume = self.previous_volume


class SonifiedDataLayerArtist(ImageLayerArtist):

    _layer_state_cls = SonifiedLayerState

    def __init__(self, view, *args, **kwargs):
        super().__init__(view, *args, **kwargs)
        self.view = view

    @property
    def audible(self):
        return self.state.audible

    @audible.setter
    def audible(self, value=None):
        if value is None:
            return
        self.state.audible = value

    def enable(self):
        if self.enabled:
            return

    def redraw(self):
        if self.view:
            # needs to run after a sonified layer is removed from the viewer and
            # then added back
            self.view.recalculate_combined_sonified_grid()

    def remove(self):
        super().remove()
        self.audible = False


class SonifiedLayerStateWidget(v.VuetifyTemplate):
    template_file = (__file__, 'layer_sonified.vue')

    def __init__(self, layer, *args, **kwargs):
        super().__init__(*args, **kwargs)



import numpy as np
from traitlets import Any, Bool, List, Unicode, observe
from astropy import units as u
from astropy.nddata import StdDevUncertainty
from specutils import Spectrum

from jdaviz.core.custom_units_and_equivs import PIX2
from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import loader_importer_registry, viewer_registry
from jdaviz.core.loaders.importers import (BaseImporterToDataCollection,
                                           SpectrumInputExtensionsMixin)
from jdaviz.core.loaders.importers.spectrum_common import _spectrum_assign_component_type
from jdaviz.core.loaders.importers.image.image import _spatial_assign_component_type
from jdaviz.core.template_mixin import (AutoTextField,
                                        SelectPluginComponent,
                                        ViewerSelectCreateNew)
from jdaviz.core.unit_conversion_utils import (check_if_unit_is_per_solid_angle,
                                               _eqv_flux_to_sb_pixel)
from jdaviz.core.user_api import ImporterUserApi


__all__ = ['Spectrum3DImporter']


@loader_importer_registry('3D Spectrum')
class Spectrum3DImporter(BaseImporterToDataCollection, SpectrumInputExtensionsMixin):
    template_file = __file__, "./spectrum3d.vue"
    parser_preference = ['fits', 'asdf', 'specutils.Spectrum']

    # Uncertainty Cube
    unc_data_label_value = Unicode().tag(sync=True)
    unc_data_label_default = Unicode().tag(sync=True)
    unc_data_label_auto = Bool(True).tag(sync=True)
    unc_data_label_invalid_msg = Unicode().tag(sync=True)

    # Uncertainty Viewer
    has_unc = Bool(False).tag(sync=True)
    unc_viewer_create_new_items = List([]).tag(sync=True)
    unc_viewer_create_new_selected = Unicode().tag(sync=True)

    unc_viewer_items = List([]).tag(sync=True)
    unc_viewer_selected = Any([]).tag(sync=True)
    unc_viewer_multiselect = Bool(True).tag(sync=True)

    unc_viewer_label_value = Unicode().tag(sync=True)
    unc_viewer_label_default = Unicode().tag(sync=True)
    unc_viewer_label_auto = Bool(True).tag(sync=True)
    unc_viewer_label_invalid_msg = Unicode().tag(sync=True)

    # Mask Cube
    mask_data_label_value = Unicode().tag(sync=True)
    mask_data_label_default = Unicode().tag(sync=True)
    mask_data_label_auto = Bool(True).tag(sync=True)
    mask_data_label_invalid_msg = Unicode().tag(sync=True)

    # Mask Viewer
    has_mask = Bool(False).tag(sync=True)
    mask_viewer_create_new_items = List([]).tag(sync=True)
    mask_viewer_create_new_selected = Unicode().tag(sync=True)

    mask_viewer_items = List([]).tag(sync=True)
    mask_viewer_selected = Any([]).tag(sync=True)
    mask_viewer_multiselect = Bool(True).tag(sync=True)

    mask_viewer_label_value = Unicode().tag(sync=True)
    mask_viewer_label_default = Unicode().tag(sync=True)
    mask_viewer_label_auto = Bool(True).tag(sync=True)
    mask_viewer_label_invalid_msg = Unicode().tag(sync=True)

    # Extraction Options
    auto_extract = Bool(True).tag(sync=True)
    function_items = List().tag(sync=True)
    function_selected = Unicode('Sum').tag(sync=True)
    # Don't load uncertainty and mask as separate cubes, e.g. for plugin results
    flux_only = Bool(False).tag(sync=True)

    # Extracted Data
    ext_data_label_value = Unicode().tag(sync=True)
    ext_data_label_default = Unicode().tag(sync=True)
    ext_data_label_auto = Bool(True).tag(sync=True)
    ext_data_label_invalid_msg = Unicode().tag(sync=True)

    # Extracted Viewer
    ext_viewer_create_new_items = List([]).tag(sync=True)
    ext_viewer_create_new_selected = Unicode().tag(sync=True)

    ext_viewer_items = List([]).tag(sync=True)
    ext_viewer_selected = Any([]).tag(sync=True)
    ext_viewer_multiselect = Bool(True).tag(sync=True)

    ext_viewer_label_value = Unicode().tag(sync=True)
    ext_viewer_label_default = Unicode().tag(sync=True)
    ext_viewer_label_auto = Bool(True).tag(sync=True)
    ext_viewer_label_invalid_msg = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def viewer_in_registry_names(supported_viewers):
            def viewer_filter(viewer):
                classes = [viewer_registry.members.get(item.get('reference')).get('cls')
                           for item in supported_viewers]
                return isinstance(viewer, tuple(classes))
            return viewer_filter

        # FLUX CUBE
        if self.default_data_label_from_resolver:
            self.data_label_default = self.default_data_label_from_resolver
        elif self.config == 'cubeviz':
            self.data_label_default = '3D Spectrum [FLUX]'
        else:
            self.data_label_default = '3D Spectrum'

        if self.config == 'cubeviz':
            self.viewer.selected = ['flux-viewer']

        # UNCERTAINTY CUBE
        self.has_unc = self.spectrum.uncertainty is not None
        self.unc_data_label = AutoTextField(self,
                                            'unc_data_label_value',
                                            'unc_data_label_default',
                                            'unc_data_label_auto',
                                            'unc_data_label_invalid_msg')

        self.unc_viewer = ViewerSelectCreateNew(self,
                                                'unc_viewer_items',
                                                'unc_viewer_selected',
                                                'unc_viewer_create_new_items',
                                                'unc_viewer_create_new_selected',
                                                'unc_viewer_label_value',
                                                'unc_viewer_label_default',
                                                'unc_viewer_label_auto',
                                                'unc_viewer_label_invalid_msg',
                                                multiselect='unc_viewer_multiselect',
                                                default_mode='empty')
        # TODO: default label separate from viewer label (so we can call it unc-viewer, etc)
        supported_viewers = [{'label': '3D Spectrum',
                              'reference': 'cubeviz-image-viewer'}]
        if self.app.config == 'deconfigged':
            self.unc_viewer_create_new_items = supported_viewers

        self.unc_viewer.add_filter(viewer_in_registry_names(supported_viewers))
        if self.config == 'cubeviz':
            self.unc_viewer.selected = ['uncert-viewer']
        else:
            self.unc_viewer.select_default()

        # MASK CUBE
        self.has_mask = self.spectrum.mask is not None
        self.mask_data_label = AutoTextField(self,
                                             'mask_data_label_value',
                                             'mask_data_label_default',
                                             'mask_data_label_auto',
                                             'mask_data_label_invalid_msg')

        self.mask_viewer = ViewerSelectCreateNew(self,
                                                 'mask_viewer_items',
                                                 'mask_viewer_selected',
                                                 'mask_viewer_create_new_items',
                                                 'mask_viewer_create_new_selected',
                                                 'mask_viewer_label_value',
                                                 'mask_viewer_label_default',
                                                 'mask_viewer_label_auto',
                                                 'mask_viewer_label_invalid_msg',
                                                 multiselect='mask_viewer_multiselect',
                                                 default_mode='empty')
        # TODO: default label separate from viewer label (so we can call it mask-viewer, etc)
        supported_viewers = [{'label': '3D Spectrum',
                              'reference': 'cubeviz-image-viewer'}]
        if self.app.config == 'deconfigged':
            self.mask_viewer_create_new_items = supported_viewers

        self.mask_viewer.add_filter(viewer_in_registry_names(supported_viewers))
        if self.config == 'cubeviz':
            self.mask_viewer.selected = []
        else:
            self.mask_viewer.selected = []

        # AUTO-EXTRACTION
        self.function = SelectPluginComponent(
            self,
            items='function_items',
            selected='function_selected',
            manual_options=['Mean', 'Min', 'Max', 'Sum']
        )

        self.ext_data_label = AutoTextField(self,
                                            'ext_data_label_value',
                                            'ext_data_label_default',
                                            'ext_data_label_auto',
                                            'ext_data_label_invalid_msg')

        self.ext_viewer = ViewerSelectCreateNew(self,
                                                'ext_viewer_items',
                                                'ext_viewer_selected',
                                                'ext_viewer_create_new_items',
                                                'ext_viewer_create_new_selected',
                                                'ext_viewer_label_value',
                                                'ext_viewer_label_default',
                                                'ext_viewer_label_auto',
                                                'ext_viewer_label_invalid_msg',
                                                multiselect='ext_viewer_multiselect',
                                                default_mode='empty')
        supported_viewers = [{'label': '1D Spectrum',
                              'reference': 'spectrum-1d-viewer'}]
        if self.app.config == 'deconfigged':
            self.ext_viewer_create_new_items = supported_viewers

        self.ext_viewer.add_filter(viewer_in_registry_names(supported_viewers))
        self.ext_viewer.select_default()

    @staticmethod
    def _get_supported_viewers():
        return [{'label': '3D Spectrum', 'reference': 'cubeviz-image-viewer'}]

    @property
    def user_api(self):
        expose = ['auto_extract', 'ext_data_label', 'ext_viewer', 'flux_only']
        if self.has_unc:
            expose += ['unc_data_label', 'unc_viewer']
        if self.has_mask:
            expose += ['mask_data_label', 'mask_viewer']
        expose += ['extension']
        if self.has_unc:
            expose += ['unc_extension']
        if self.has_mask:
            expose += ['mask_extension']
        return ImporterUserApi(self, expose)

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'cubeviz'):
            # NOTE: temporary during deconfig process
            return False
        try:
            if self.spectrum.flux.ndim != 3:
                return False
        except Exception:
            return False

        try:
            self.output
        except Exception:
            return False
        return True

    @observe('data_label_value', 'function_selected')
    def _data_label_changed(self, msg={}):
        base = self.data_label_value.strip('[FLUX]').strip()
        self.ext_data_label_default = f"{base} ({self.function_selected.lower()})"
        self.unc_data_label_default = f"{base} [UNC]"
        self.mask_data_label_default = f"{base} [MASK]"

    @property
    def supported_flux_ndim(self):
        return 3

    @property
    def default_spectral_axis_index(self):
        return 0

    @property
    def output(self):
        sp = self.spectrum

        # convert flux and uncertainty to per-pix2 if input is not a surface brightness
        if not check_if_unit_is_per_solid_angle(sp.flux.unit):
            target_flux_unit = sp.flux.unit / PIX2
        elif check_if_unit_is_per_solid_angle(sp.flux.unit, return_unit=True) == "spaxel":
            # We need to convert spaxel to pix2, since spaxel isn't fully supported by astropy
            # This is horribly ugly but just multiplying by u.Unit("spaxel") doesn't work
            target_flux_unit = sp.flux.unit * u.Unit('spaxel') / PIX2
        else:
            target_flux_unit = sp.flux.unit

        # handle scale factors when they are included in the unit
        if not np.isclose(target_flux_unit.scale, 1, rtol=1e-5):
            target_flux_unit = target_flux_unit / target_flux_unit.scale

        if target_flux_unit == sp.flux.unit:
            return sp

        # In the case of a np.array, it is loaded into a Spectrum without WCS,
        # so specutils creates a SpectralGWCS by default. We need this WCS later on
        # for spatial region handling, so we store a copy of that WCS here.
        if '_orig_spatial_wcs' not in sp.meta:
            sp.meta['_orig_spatial_wcs'] = sp.wcs

        return sp.with_flux_unit(target_flux_unit, equivalencies=_eqv_flux_to_sb_pixel())

    def __call__(self):
        # get a copy of all requested data-labels before additional data entries changes defaults
        data_label = self.data_label_value
        unc_data_label = self.unc_data_label_value
        mask_data_label = self.mask_data_label_value
        ext_data_label = self.ext_data_label_value

        super().__call__()
        # TODO: this will need to be removed when removing restriction of a single flux cube
        if not getattr(self.app._jdaviz_helper, '_loaded_flux_cube', None):
            self.app._jdaviz_helper._loaded_flux_cube = self.app.data_collection[data_label]

        if self.has_unc and not self.flux_only:
            # TODO: detect if uncertainty exists and hide section from UI
            uncert = Spectrum(spectral_axis=self.output.spectral_axis,
                              flux=self.output.uncertainty.represent_as(StdDevUncertainty).quantity,
                              wcs=self.output.wcs,
                              meta=self.output.meta,
                              spectral_axis_index=self.output.spectral_axis_index)
            self.add_to_data_collection(uncert,
                                        unc_data_label,
                                        viewer_select=self.unc_viewer)
            # TODO: this will need to be removed when removing restriction of a single flux cube
            self.app._jdaviz_helper._loaded_uncert_cube = self.app.data_collection[unc_data_label]

        if self.has_mask and not self.flux_only:
            mask = Spectrum(spectral_axis=self.output.spectral_axis,
                            flux=self.output.mask * u.dimensionless_unscaled,
                            wcs=self.output.wcs,
                            meta=self.output.meta,
                            spectral_axis_index=self.output.spectral_axis_index)
            self.add_to_data_collection(mask,
                                        mask_data_label,
                                        viewer_select=self.mask_viewer)
            # TODO: this will need to be removed when removing restriction of a single flux cube
            self.app._jdaviz_helper._loaded_mask_cube = self.app.data_collection[mask_data_label]

        if not self.auto_extract:
            return

        try:
            spext = self.app.get_tray_item_from_name('spectral-extraction-3d')
            ext = spext._extract_in_new_instance(dataset=data_label,
                                                 function=self.function.selected,
                                                 auto_update=False,
                                                 add_data=False)
            # we'll add the data manually instead of through add_results_from_plugin
            # but still want to preserve the plugin metadata
            ext.meta['plugin'] = spext._plugin_name
        except Exception as e:
            ext = None
            msg = SnackbarMessage(
                "Automatic spectrum extraction failed. See the 3D spectral extraction"
                " plugin to perform a custom extraction",
                color='error', sender=self, timeout=10000, traceback=e)
        else:
            msg = SnackbarMessage(
                "The extracted 1D spectrum was generated automatically."
                " See the 3D spectral extraction plugin for details or to"
                " perform a custom extraction.",
                color='warning', sender=self, timeout=10000)
        self.app.hub.broadcast(msg)

        if ext is not None:
            self.add_to_data_collection(ext, ext_data_label, viewer_select=self.ext_viewer)

    def assign_component_type(self, comp_id, comp, units, physical_type):
        comp_type = _spatial_assign_component_type(comp_id, comp, units, physical_type)
        return _spectrum_assign_component_type(comp_id, comp, units, comp_type)



import numpy as np
import pytest
from astropy import units as u
from astropy.nddata import InverseVariance
from specutils import Spectrum

from jdaviz.core.custom_units_and_equivs import SPEC_PHOTON_FLUX_DENSITY_UNITS


# On failure, should not crash; essentially a no-op.
@pytest.mark.parametrize(
    ('new_spectral_axis', 'new_flux', 'expected_spectral_axis', 'expected_flux'),
    [("fail", "erg / (s cm2 Angstrom)", "Angstrom", "erg / (s cm2 Angstrom)"),
     ("None", "fail", "Angstrom", "Jy"),
     ("micron", "fail", "micron", "Jy")])
def test_value_error_exception(specviz_helper, spectrum1d, new_spectral_axis, new_flux,
                               expected_spectral_axis, expected_flux):
    specviz_helper.load_data(spectrum1d, data_label="Test 1D Spectrum")
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')
    plg = specviz_helper.plugins["Unit Conversion"]

    try:
        plg.spectral_unit = new_spectral_axis
    except ValueError as e:
        if "reverting selection to" not in repr(e):
            raise
    try:
        plg._obj.flux_unit_selected = new_flux
    except ValueError as e:
        if "reverting selection to" not in repr(e):
            raise

    assert len(specviz_helper.app.data_collection) == 1
    assert u.Unit(viewer.state.x_display_unit) == u.Unit(expected_spectral_axis)
    assert u.Unit(viewer.state.y_display_unit) == u.Unit(expected_flux)


def test_initialize_specviz_sb(specviz_helper, spectrum1d):
    spec_sb = Spectrum(spectrum1d.flux/u.sr, spectrum1d.spectral_axis)
    specviz_helper.load_data(spec_sb, data_label="Test 1D Spectrum")
    plg = specviz_helper.plugins["Unit Conversion"]
    assert plg._obj.flux_unit == "Jy"
    assert plg._obj.spectral_y_type == "Surface Brightness"
    assert plg._obj.angle_unit == "sr"


@pytest.mark.parametrize('uncert', (False, True))
def test_conv_wave_only(specviz_helper, spectrum1d, uncert):
    if uncert is False:
        spectrum1d.uncertainty = None
    specviz_helper.load_data(spectrum1d, data_label="Test 1D Spectrum")

    viewer = specviz_helper.app.get_viewer('spectrum-viewer')
    plg = specviz_helper.plugins["Unit Conversion"]
    new_spectral_axis = "micron"
    plg.spectral_unit = new_spectral_axis

    assert len(specviz_helper.app.data_collection) == 1
    assert u.Unit(viewer.state.x_display_unit) == u.Unit(new_spectral_axis)
    assert u.Unit(viewer.state.y_display_unit) == u.Unit('Jy')


@pytest.mark.parametrize('uncert', (False, True))
def test_conv_flux_only(specviz_helper, spectrum1d, uncert):
    if uncert is False:
        spectrum1d.uncertainty = None
    specviz_helper.load_data(spectrum1d, data_label="Test 1D Spectrum")

    viewer = specviz_helper.app.get_viewer('spectrum-viewer')
    plg = specviz_helper.plugins["Unit Conversion"]
    new_flux = "erg / (s cm2 Angstrom)"
    plg._obj.flux_unit_selected = new_flux

    assert len(specviz_helper.app.data_collection) == 1
    assert u.Unit(viewer.state.x_display_unit) == u.Unit('Angstrom')
    assert u.Unit(viewer.state.y_display_unit) == u.Unit(new_flux)


@pytest.mark.parametrize('uncert', (False, True))
def test_conv_wave_flux(specviz_helper, spectrum1d, uncert):
    if uncert is False:
        spectrum1d.uncertainty = None
    specviz_helper.load_data(spectrum1d, data_label="Test 1D Spectrum")

    viewer = specviz_helper.app.get_viewer('spectrum-viewer')
    plg = specviz_helper.plugins["Unit Conversion"]
    new_spectral_axis = "micron"
    new_flux = "erg / (s cm2 Angstrom)"
    plg.spectral_unit = new_spectral_axis
    plg._obj.flux_unit_selected = new_flux

    assert len(specviz_helper.app.data_collection) == 1
    assert u.Unit(viewer.state.x_display_unit) == u.Unit(new_spectral_axis)
    assert u.Unit(viewer.state.y_display_unit) == u.Unit(new_flux)


def test_conv_no_data(specviz_helper, spectrum1d):
    """plugin unit selections won't have valid choices yet, preventing
    attempting to set display units."""
    # spectrum not load is in Flux units, sb_unit and flux_unit
    # should be enabled, spectral_y_type should not be
    plg = specviz_helper.plugins["Unit Conversion"]
    with pytest.raises(ValueError, match="could not find match in valid x display units"):
        plg.spectral_unit = "micron"
    assert len(specviz_helper.app.data_collection) == 0

    specviz_helper.load_data(spectrum1d, data_label="Test 1D Spectrum")

    # make sure we don't expose translations in Specviz
    assert hasattr(plg, 'flux_unit')
    assert hasattr(plg, 'angle_unit')
    assert not hasattr(plg, 'spectral_y_type')


def test_non_stddev_uncertainty(specviz_helper):
    flux = np.ones(10) * u.Jy
    stddev = 0.1
    var = stddev ** 2
    inv_var = np.ones(len(flux)) / var
    wavelength = np.linspace(1, 5, len(flux)) * u.um
    spec = Spectrum(
        flux,
        uncertainty=InverseVariance(inv_var),
        spectral_axis=wavelength
    )

    specviz_helper.load_data(spec)

    po = specviz_helper.plugins['Plot Options']
    po.uncertainty_visible = True

    # check that the stddev uncertainties are drawn:
    viewer = specviz_helper.app.get_viewer('spectrum-viewer')
    np.testing.assert_allclose(
        np.abs(viewer.figure.marks[-1].y - viewer.figure.marks[-1].y.mean(0)),
        stddev
    )


@pytest.mark.parametrize("flux_unit, expected_choices", [(u.count, ['ct']),
                                                         (u.Jy, SPEC_PHOTON_FLUX_DENSITY_UNITS),
                                                         (u.nJy, SPEC_PHOTON_FLUX_DENSITY_UNITS + ['nJy'])])  # noqa
def test_flux_unit_choices(specviz_helper, flux_unit, expected_choices):
    """
    Test that cubes loaded with various flux units have the expected default
    flux unit selection in the unit conversion plugin, and that the list of
    convertable flux units in the dropdown is correct.
    """

    spec = Spectrum([1, 2, 3] * flux_unit, [4, 5, 6] * u.um)
    specviz_helper.load_data(spec)

    uc_plg = specviz_helper.plugins['Unit Conversion']

    assert uc_plg.flux_unit.selected == flux_unit.to_string()
    assert uc_plg.flux_unit.choices == expected_choices


def test_mosviz_viewer_mouseover_flux(specviz2d_helper):
    data = np.zeros((5, 10))
    data[3] = np.arange(10)
    spectrum2d = Spectrum(flux=data*u.MJy, spectral_axis=data[3]*u.um)

    specviz2d_helper.load_data(spectrum2d)
    viewer = specviz2d_helper.app.get_viewer('spectrum-viewer')
    plg = specviz2d_helper.plugins["Unit Conversion"]

    # make sure we don't expose angle, sb, nor spectral-y units when native
    # units are in flux
    assert hasattr(plg, 'flux_unit')
    assert not hasattr(plg, 'angle_unit')
    assert not hasattr(plg, 'sb_unit')
    assert not hasattr(plg, 'spectral_y_type')

    label_mouseover = specviz2d_helper._coords_info
    label_mouseover._viewer_mouse_event(viewer,
                                        {'event': 'mousemove',
                                         'domain': {'x': 5, 'y': 3}})

    assert label_mouseover.as_text() == ('Cursor 5.00000e+00, 3.00000e+00',
                                         'Wave 5.00000e+00 um (5 pix)',
                                         'Flux 5.00000e+00 MJy')

    assert viewer.axis_y.label == 'Flux[MJy]'

    plg._obj.flux_unit_selected = 'Jy'
    # ensure axis label updates when flux unit is changed
    assert viewer.axis_y.label == 'Flux[Jy]'

    label_mouseover._viewer_mouse_event(viewer,
                                        {'event': 'mousemove',
                                         'domain': {'x': 5, 'y': 3}})

    assert label_mouseover.as_text() == ('Cursor 5.00000e+00, 3.00000e+00',
                                         'Wave 5.00000e+00 um (5 pix)',
                                         'Flux 5.00000e+06 Jy')

    # test mouseover when spectral density equivalencies are required for conversion
    plg._obj.flux_unit_selected = 'erg / (Angstrom s cm2)'
    assert viewer.axis_y.label == 'Flux[erg / (Angstrom s cm2)]'

    assert label_mouseover.as_text() == ('Cursor 5.00000e+00, 3.00000e+00',
                                         'Wave 5.00000e+00 um (5 pix)',
                                         'Flux 5.99585e-08 erg / (Angstrom s cm2)')


def test_mosviz_viewer_mouseover_sb(specviz2d_helper):
    data = np.zeros((5, 10))
    data[3] = np.arange(10)
    spectrum2d = Spectrum(flux=data*u.MJy/u.sr, spectral_axis=data[3]*u.um)

    specviz2d_helper.load_data(spectrum2d)
    spectrum_viewer = specviz2d_helper.app.get_viewer("spectrum-viewer")
    spectrum2d_viewer = specviz2d_helper.app.get_viewer('spectrum-2d-viewer')
    plg = specviz2d_helper.plugins["Unit Conversion"]

    # make sure we don't expose angle, sb, nor spectral-y units when native
    # units are in flux
    assert hasattr(plg, 'flux_unit')
    assert hasattr(plg, 'angle_unit')
    assert hasattr(plg, 'sb_unit')
    assert not hasattr(plg, 'spectral_y_type')

    label_mouseover = specviz2d_helper.app.session.application._tools['g-coords-info']
    label_mouseover._viewer_mouse_event(spectrum_viewer,
                                        {'event': 'mousemove',
                                         'domain': {'x': 5, 'y': 3}})
    assert label_mouseover.as_text() == ('Cursor 5.00000e+00, 3.00000e+00',
                                         'Wave 5.00000e+00 um (5 pix)',
                                         'Flux 5.00000e+00 MJy / sr')

    label_mouseover._viewer_mouse_event(spectrum2d_viewer, {"event": "mousemove",
                                                            "domain": {"x": 5, "y": 3}})
    output2d = label_mouseover.as_text()
    expected2d = ("Pixel x=05.0 y=03.0 Value +5.00000e+00 MJy / sr",
                  "Wave 5.00000e+00 um",
                  '')
    assert output2d == expected2d

    assert spectrum_viewer.axis_y.label == 'Surface Brightness[MJy / sr]'

    plg._obj.flux_unit_selected = 'Jy'
    assert spectrum_viewer.axis_y.label == 'Surface Brightness[Jy / sr]'

    label_mouseover._viewer_mouse_event(spectrum_viewer,
                                        {'event': 'mousemove',
                                         'domain': {'x': 5, 'y': 3}})
    output1d = label_mouseover.as_text()
    expected1d = ("Cursor 5.00000e+00, 3.00000e+00 Value +5.00000e+06 Jy / sr",
                  "Wave 5.00000e+00 um (5 pix)",
                  "Flux 5.00000e+06 Jy / sr")
    assert output1d == expected1d

    label_mouseover._viewer_mouse_event(spectrum2d_viewer, {"event": "mousemove",
                                                            "domain": {"x": 5, "y": 3}})
    output2d = label_mouseover.as_text()
    expected2d = ("Pixel x=05.0 y=03.0 Value +5.00000e+06 Jy / sr",
                  "Wave 5.00000e+00 um",
                  '')
    assert output2d == expected2d

    # test mouseover when spectral density equivalencies are required for conversion
    plg._obj.flux_unit_selected = 'erg / (Angstrom s cm2)'
    assert spectrum_viewer.axis_y.label == 'Surface Brightness[erg / (Angstrom s sr cm2)]'

    label_mouseover._viewer_mouse_event(spectrum_viewer,
                                        {'event': 'mousemove',
                                         'domain': {'x': 5, 'y': 3}})
    output1d = label_mouseover.as_text()
    expected1d = ("Cursor 5.00000e+00, 3.00000e+00 Value +5.99585e-08 erg / (Angstrom s sr cm2)",
                  "Wave 5.00000e+00 um (5 pix)",
                  "Flux 5.99585e-08 erg / (Angstrom s sr cm2)")
    assert output1d == expected1d

    label_mouseover._viewer_mouse_event(spectrum2d_viewer, {"event": "mousemove",
                                                            "domain": {"x": 5, "y": 3}})
    output2d = label_mouseover.as_text()
    expected2d = ("Pixel x=05.0 y=03.0 Value +5.99585e-08 erg / (Angstrom s sr cm2)",
                  "Wave 5.00000e+00 um",
                  '')
    assert output2d == expected2d


def test_image_deconfigged(deconfigged_helper, image_nddata_wcs):
    """
    Test that the unit conversion plugin works in deconfigged mode.
    """
    deconfigged_helper.load(image_nddata_wcs, format='Image', data_label="Test Image")
    plg = deconfigged_helper.plugins["Unit Conversion"]

    viewer = deconfigged_helper.viewers['Image']
    label_mouseover = deconfigged_helper.app.session.application._tools['g-coords-info']

    assert plg.flux_unit == "Jy"

    label_mouseover._viewer_mouse_event(viewer._obj.glue_viewer,
                                        {'event': 'mousemove',
                                         'domain': {'x': 1, 'y': 1}})
    assert label_mouseover.as_text() == ('Pixel x=01.0 y=01.0 Value +1.00000e+00 Jy',
                                         'World 22h30m04.7961s -20d49m58.9990s (ICRS)',
                                         '337.5199835909 -20.8330552820 (deg)')

    # Check that the units can be changed
    plg.flux_unit = "MJy"

    assert plg.flux_unit == "MJy"

    label_mouseover._viewer_mouse_event(viewer._obj.glue_viewer,
                                        {'event': 'mousemove',
                                         'domain': {'x': 1, 'y': 1}})
    assert label_mouseover.as_text() == ('Pixel x=01.0 y=01.0 Value +1.00000e-06 MJy',
                                         'World 22h30m04.7961s -20d49m58.9990s (ICRS)',
                                         '337.5199835909 -20.8330552820 (deg)')


def test_data_unload_reload(specviz2d_helper):
    """
    Test that when data is loaded (which sets ths initial unit selection
    for the unit conversion plugin), then removed, and a new dataset
    with different units is loaded, that the spectrum viewer is in the
    original units set by the first dataset. This behavior aligns with the
    idea that UC plugin selections should be app-wide settings.
    """
    # load initial data in MJy, um
    data = np.zeros((5, 10))
    data[3] = np.arange(10)
    spectrum2d = Spectrum(flux=data*u.MJy, spectral_axis=data[3]*u.um)

    specviz2d_helper.load_data(spectrum2d)

    # remove data from viewer
    viewer = specviz2d_helper.app.get_viewer('spectrum-viewer')
    dm = viewer.data_menu
    dm.remove_from_viewer()

    # make sure initial unit setting matches data units in both
    # plugin and on the spectrum viewer axes
    uc = specviz2d_helper.plugins['Unit Conversion']
    assert uc.flux_unit.selected == viewer.state.y_display_unit == 'MJy'
    assert uc.spectral_unit == viewer.state.x_display_unit == 'um'

    # re-load data with new units into the spectrum viewer
    spectrum2d_new = Spectrum(flux=data*u.Jy/u.sr, spectral_axis=data[3]*u.nm)
    specviz2d_helper.load_data(spectrum2d_new)

    # make sure newly loaded data did not change unit conversion selection,
    # and that the plugin traitlets and viewer attributes still match the
    # initially loaded unit
    uc = specviz2d_helper.plugins['Unit Conversion']
    assert uc.flux_unit.selected == viewer.state.y_display_unit == 'MJy'
    assert uc.spectral_unit == viewer.state.x_display_unit == 'um'



"""
Comprehensive test coverage for SubsetImporter.

Test coverage includes:
- Validation for Regions and SpectralRegion inputs
- Label validation and generation
- Import functionality with Subset Tools plugin
- Edge cases and error handling
"""
import pytest
from regions import CirclePixelRegion, PixCoord, Regions
from specutils import SpectralRegion

import astropy.units as u

from jdaviz.core.loaders.importers.subset.subset import SubsetImporter


@pytest.fixture
def circle_region():
    """
    Create a simple CirclePixelRegion for testing.
    """
    return CirclePixelRegion(center=PixCoord(x=10, y=20), radius=5)


@pytest.fixture
def regions_input(circle_region):
    """
    Create a Regions object for testing.
    """
    return Regions([circle_region])


@pytest.fixture
def spectral_region():
    """
    Create a SpectralRegion for testing.
    """
    return SpectralRegion(5000 * u.AA, 6000 * u.AA)


class TestSubsetImporter:

    @pytest.fixture(autouse=True)
    def _setup(self, deconfigged_helper, spectrum1d):
        deconfigged_helper.load(spectrum1d, format='1D Spectrum')
        self.dcf_helper = deconfigged_helper
        self.app = deconfigged_helper.app

    def generate_importer(self, input_data):
        return SubsetImporter(app=self.app,
                              resolver=None,
                              parser=None,
                              input=input_data)

    def test_label_generation_and_validity(self, regions_input):
        """
        Test label things and subsequent valid calls/error messages.
        """
        importer = self.generate_importer(regions_input)

        # Test that default label (Subset N) passes validation.
        importer.subset_label_value = 'Subset 1'
        importer._on_label_changed()

        # Default label should be valid
        assert importer.subset_label_invalid_msg == ''
        assert importer.import_disabled is False

        # Label validation works correctly.
        for label_value in ('', '    '):
            importer.subset_label_value = label_value
            importer._on_label_changed()
            assert importer.subset_label_invalid_msg == 'subset_label must be provided'

        # Test that the label_invalid_msg is set for an invalid subset label.
        importer.subset_label_value = 'Subset 2'
        importer._on_label_changed()

        # Default label should be invalid
        error_msg = ("invalid subset_label: "
                     "The pattern 'Subset N' "
                     "is reserved for auto-generated labels")
        assert importer.subset_label_invalid_msg == error_msg

        # check import_disabled updates correctly
        assert importer.import_disabled is True

        # test that calling with invalid label raises ValueError
        with pytest.raises(ValueError, match=error_msg):
            importer()

    def test_is_valid(self, regions_input, spectral_region):
        """
        Test is_valid for various scenarios.
        """
        importer = self.generate_importer(regions_input)

        assert importer.is_valid is True
        assert importer.default_plugin == 'Subset Tools'

        importer = self.generate_importer(spectral_region)
        assert importer.is_valid is True

        importer = self.generate_importer('not a region')
        assert importer.is_valid is False

    def test_label_default_updates_with_subset_count(self, regions_input, spectral_region):
        """
        Test that default label updates based on subset count.
        """
        importer = self.generate_importer(regions_input)
        # Should have default label set
        assert importer.subset_label_default == 'Subset 1'
        assert importer.subset_label_auto is True

        # subset_label should be set
        assert hasattr(importer, 'subset_label')
        assert importer.subset_label is not None

        # Create a subset first
        subset_plugin = self.dcf_helper.plugins['Subset Tools']
        subset_plugin.import_region(spectral_region)

        # Now create importer and check default updates
        importer = self.generate_importer(regions_input)

        importer.subset_label_value = 'Test'
        importer._on_label_changed()

        # Should be Subset 2 since we created one subset
        assert importer.subset_label_default == 'Subset 2'

    def test_call_creates_subset(self, spectral_region):
        """
        Test that calling importer creates a subset.
        """
        importer = self.generate_importer(spectral_region)

        # Ensure valid state
        importer.subset_label_value = 'Subset 1'
        importer.subset_label_invalid_msg = ''

        # Call should not raise
        initial_subset_count = len(self.app.data_collection.subset_groups)
        importer()

        # Should have created a new subset
        assert len(self.app.data_collection.subset_groups) >= initial_subset_count



from traitlets import Any, Bool, List, Unicode, observe

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import loader_importer_registry, viewer_registry
from jdaviz.core.loaders.importers import (BaseImporterToDataCollection,
                                           SpectrumInputExtensionsMixin,
                                           _spectrum_assign_component_type)
from jdaviz.core.template_mixin import (AutoTextField,
                                        ViewerSelectCreateNew)
from jdaviz.core.user_api import ImporterUserApi


__all__ = ['Spectrum2DImporter']


@loader_importer_registry('2D Spectrum')
class Spectrum2DImporter(BaseImporterToDataCollection, SpectrumInputExtensionsMixin):
    template_file = __file__, "./spectrum2d.vue"
    parser_preference = ['fits', 'asdf', 'specutils.Spectrum']

    auto_extract = Bool(True).tag(sync=True)

    ext_data_label_value = Unicode().tag(sync=True)
    ext_data_label_default = Unicode().tag(sync=True)
    ext_data_label_auto = Bool(True).tag(sync=True)
    ext_data_label_invalid_msg = Unicode().tag(sync=True)

    ext_viewer_create_new_items = List([]).tag(sync=True)
    ext_viewer_create_new_selected = Unicode().tag(sync=True)
    # No uncertainty viewer for 2D
    has_unc = Bool(False).tag(sync=True)

    ext_viewer_items = List([]).tag(sync=True)
    ext_viewer_selected = Any([]).tag(sync=True)
    ext_viewer_multiselect = Bool(True).tag(sync=True)

    ext_viewer_label_value = Unicode().tag(sync=True)
    ext_viewer_label_default = Unicode().tag(sync=True)
    ext_viewer_label_auto = Bool(True).tag(sync=True)
    ext_viewer_label_invalid_msg = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.default_data_label_from_resolver:
            self.data_label_default = self.default_data_label_from_resolver
        elif self.app.config == 'specviz2d':
            self.data_label_default = '2D Spectrum'

        self.ext_data_label = AutoTextField(self,
                                            'ext_data_label_value',
                                            'ext_data_label_default',
                                            'ext_data_label_auto',
                                            'ext_data_label_invalid_msg')

        self.ext_viewer = ViewerSelectCreateNew(self,
                                                'ext_viewer_items',
                                                'ext_viewer_selected',
                                                'ext_viewer_create_new_items',
                                                'ext_viewer_create_new_selected',
                                                'ext_viewer_label_value',
                                                'ext_viewer_label_default',
                                                'ext_viewer_label_auto',
                                                'ext_viewer_label_invalid_msg',
                                                multiselect='ext_viewer_multiselect',
                                                default_mode='empty')
        supported_viewers = [{'label': '1D Spectrum',
                              'reference': 'spectrum-1d-viewer'}]
        if self.app.config == 'deconfigged':
            self.ext_viewer_create_new_items = supported_viewers

        def viewer_in_registry_names(viewer):
            classes = [viewer_registry.members.get(item.get('reference')).get('cls')
                       for item in supported_viewers]
            return isinstance(viewer, tuple(classes))
        self.ext_viewer.add_filter(viewer_in_registry_names)
        self.ext_viewer.select_default()

    @staticmethod
    def _get_supported_viewers():
        return [{'label': '2D Spectrum', 'reference': 'spectrum-2d-viewer'}]

    @property
    def user_api(self):
        expose = ['auto_extract', 'ext_data_label', 'ext_viewer',
                  'extension', 'unc_extension', 'mask_extension']
        return ImporterUserApi(self, expose)

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'specviz2d'):
            # NOTE: temporary during deconfig process
            return False
        try:
            if self.spectrum.flux.ndim != 2:
                return False
        except Exception:
            return False
        try:
            self.output
        except Exception:
            return False
        return True

    @observe('data_label_value')
    def _data_label_changed(self, msg={}):
        self.ext_data_label_default = f"{self.data_label_value} (auto-ext)"

    @property
    def output(self):
        return self.spectrum

    def assign_component_type(self, comp_id, comp, units, physical_type):
        return _spectrum_assign_component_type(comp_id, comp, units, physical_type)

    def __call__(self):
        # get a copy of both of these before additional data entries changes defaults
        data_label = self.data_label_value
        ext_data_label = self.ext_data_label_value

        super().__call__()

        if not self.auto_extract:
            return

        try:
            spext = self.app.get_tray_item_from_name('spectral-extraction-2d')
            ext = spext._extract_in_new_instance(dataset=data_label,
                                                 add_data=False)
        except Exception as e:
            ext = None
            msg = SnackbarMessage(
                "Automatic spectrum extraction failed. See the 2D spectral extraction"
                " plugin to perform a custom extraction",
                color='error', sender=self, timeout=10000, traceback=e)
        else:
            msg = SnackbarMessage(
                "The extracted 1D spectrum was generated automatically."
                " See the 2D spectral extraction plugin for details or to"
                " perform a custom extraction.",
                color='warning', sender=self, timeout=10000)
        self.app.hub.broadcast(msg)

        if ext is not None:
            self.add_to_data_collection(ext, ext_data_label, viewer_select=self.ext_viewer)



import time

from astropy.nddata import CCDData
import astropy.units as u
from astropy.wcs import WCS
import numpy as np
from scipy.special import erf
from traitlets import Bool, List, Unicode, observe

from jdaviz.core.custom_traitlets import IntHandleEmpty, FloatHandleEmpty
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, DatasetSelectMixin,
                                        SpectralSubsetSelectMixin, with_spinner,
                                        AddResultsMixin)
from jdaviz.core.user_api import PluginUserApi
from jdaviz.core.events import SnackbarMessage, AddDataMessage
from jdaviz.configs.cubeviz.plugins.cube_listener import CubeListenerData, INT_MAX
from jdaviz.core.sonified_layers import SonifiedLayerState

__all__ = ['SonifyData']

try:
    import strauss  # noqa
    import sounddevice as sd
except ImportError:
    class Empty:
        pass
    sd = Empty()
    sd.default = Empty()
    sd.default.device = [-1, -1]
    _has_strauss = False
else:
    _has_strauss = True


@tray_registry('cubeviz-sonify-data', label="Sonify Data",
               category="data:analysis")
class SonifyData(PluginTemplateMixin, DatasetSelectMixin, SpectralSubsetSelectMixin,
                 AddResultsMixin):
    """
    See the :ref:`Sonify Data Plugin Documentation <cubeviz-sonify-data>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    """
    template_file = __file__, "sonify_data.vue"

    # Removing UI option to vary these for now
    sample_rate = 44100  # IntHandleEmpty(44100).tag(sync=True)
    buffer_size = 2048  # IntHandleEmpty(2048).tag(sync=True)
    assidx = FloatHandleEmpty(2.5).tag(sync=True)
    ssvidx = FloatHandleEmpty(0.65).tag(sync=True)
    eln = Bool(True).tag(sync=True)
    audfrqmin = FloatHandleEmpty(50).tag(sync=True)
    audfrqmax = FloatHandleEmpty(1000).tag(sync=True)
    use_pccut = Bool(True).tag(sync=True)
    pccut = IntHandleEmpty(20).tag(sync=True)
    volume = IntHandleEmpty(100).tag(sync=True)
    stream_active = Bool(True).tag(sync=True)
    has_strauss = Bool(_has_strauss).tag(sync=True)

    # TODO: can we referesh the list, so sounddevices are up-to-date when dropdown clicked?
    sound_devices_items = List().tag(sync=True)
    sound_devices_selected = Unicode('').tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._plugin_description = 'Sonify a data cube'
        self.docs_description = 'Sonify a data cube using the Strauss package.'
        if not self.has_strauss or sd.default.device[1] < 0:
            self.disabled_msg = ('To use Sonify Data, install strauss and restart Jdaviz. You '
                                 'can do this by running pip install strauss in the command'
                                 ' line and then launching Jdaviz. Currently, this plugin only'
                                 ' works on devices with valid sound output.')

        else:
            self.sound_device_indexes = None
            self.refresh_device_list()

        self.add_results.viewer.add_filter('is_image_viewer')
        self.add_to_viewer_selected = 'flux-viewer'
        self.sonified_cube = None
        self.sonified_viewers = []
        self.sonification_wl_ranges = None
        self.sonification_wl_unit = None
        self.stream = None
        # Dictionary that contains keys with UUIDs for each
        # sonified data layer. The value of each key is another dictionary containing
        # coordinates as keys and arrays representing sounds as the value.
        self.data_lookup = {}

        self._update_label_default(None)

        self.hub.subscribe(self, AddDataMessage,
                           handler=self._data_added_to_viewer)

        if self.config == "deconfigged":
            self.observe_traitlets_for_relevancy(traitlets_to_observe=['dataset_items'])

    @property
    def user_api(self):
        expose = ['sonify_cube']
        return PluginUserApi(self, expose)

    @observe('results_label_invalid_msg')
    def _update_label_default(self, event):
        """
        Update default label when a new sonification is added to the viewer.
        """
        # Modify default label to avoid vue error from re-using label
        self.results_label_default = self.app.return_unique_name('Sonified data', typ='data')

    def _data_added_to_viewer(self, msg):
        # Keep track of the volume attribute for each layer.
        msg.viewer.sonified_layers_enabled = []
        for layer in msg.viewer.state.layers:
            if not isinstance(layer, SonifiedLayerState):
                continue

            # Add viewer to sonified_viewers if it isn't there already
            if msg.viewer not in self.sonified_viewers:
                self.sonified_viewers.append(msg.viewer)

            # Find layer, add volume check to dictionary and add callback to volume changing and
            # audible changing
            msg.viewer.layer_volume[layer.layer.label] = layer.volume
            msg.viewer.sonified_layers_enabled += ([layer.layer.label] if
                                                   getattr(layer, 'audible', False) else [])  # noqa

            layer.remove_callback('volume', msg.viewer.recalculate_combined_sonified_grid)
            layer.remove_callback('audible', msg.viewer.recalculate_combined_sonified_grid)
            layer.add_callback('volume', msg.viewer.recalculate_combined_sonified_grid)
            layer.add_callback('audible', msg.viewer.recalculate_combined_sonified_grid)

    def start_stream(self):
        if self.stream and not self.stream.closed:
            self.stream.start()

    def stop_stream(self):
        if self.stream and not self.stream.closed:
            self.stream.stop()

    def sonify_cube(self):
        """
        Create a sonified grid in the flux viewer so that sound plays when mousing over the viewer.
        You can select the device index for audio output and also use a spectral subset to set a
        range for sonification.
        """
        t0 = time.time()
        if self.disabled_msg:
            raise ValueError('Unable to sonify cube')

        # Get index of selected device
        selected_device_index = self.sound_device_indexes[self.sound_devices_selected]

        # Apply spectral subset bounds
        if self.spectral_subset_selected != self.spectral_subset.default_text:
            display_unit = self.spectrum_viewer.state.x_display_unit
            min_wavelength = self.spectral_subset.selected_obj.lower.to_value(u.Unit(display_unit))
            max_wavelength = self.spectral_subset.selected_obj.upper.to_value(u.Unit(display_unit))
            self.sonification_wl_ranges = ([min_wavelength, max_wavelength],)
            self.sonification_wl_unit = display_unit

        # Ensure the current spectral region bounds are up-to-date at render time
        self.update_wavelength_range(None)

        previous_label = self.results_label

        # generate the sonified cube
        sonified_cube = self.get_sonified_cube(self.sample_rate, self.buffer_size,
                                               selected_device_index, self.assidx,
                                               self.ssvidx, self.pccut, self.audfrqmin,
                                               self.audfrqmax, self.eln,
                                               self.use_pccut, self.results_label)
        sonified_cube.meta['Sonified'] = True

        # For now, this still initially adds the sonified data to flux-viewer
        self.add_results.add_results_from_plugin(sonified_cube, replace=False)
        self.add_results.viewer.selected_obj.recalculate_combined_sonified_grid()

        t1 = time.time()
        msg = SnackbarMessage(f"'{previous_label}' sonified successfully in {t1-t0} seconds.",
                              color='success',
                              sender=self)
        self.app.hub.broadcast(msg)

    def get_sonified_cube(self, sample_rate, buffer_size, device, assidx, ssvidx,
                          pccut, audfrqmin, audfrqmax, eln, use_pccut, results_label):
        spectrum = self.dataset.selected_obj
        wlens = spectrum.wavelength.to('m').value
        flux = spectrum.flux.value
        self.sample_rate = sample_rate
        self.buffer_size = buffer_size

        if self.sonification_wl_ranges:
            wdx = np.zeros(wlens.size).astype(bool)
            for r in self.sonification_wl_ranges:
                # index just the spectral subregion
                wdx = np.logical_or(wdx,
                                    np.logical_and(wlens >= r[0].to_value(u.m),
                                                   wlens <= r[1].to_value(u.m)))
            wlens = wlens[wdx]
            flux_slices = [slice(None),]*3
            flux_slices[spectrum.spectral_axis_index] = wdx
            flux = flux[*flux_slices]

        pc_cube = np.percentile(np.nan_to_num(flux), np.clip(pccut, 0, 99),
                                axis=spectrum.spectral_axis_index)

        # clip zeros and remove NaNs
        clipped_arr = np.nan_to_num(np.clip(flux, 0, np.inf), copy=False)

        # make a rough white-light image from the clipped array
        whitelight = np.expand_dims(clipped_arr.sum(spectrum.spectral_axis_index),
                                    axis=spectrum.spectral_axis_index)

        if use_pccut:
            # subtract any percentile cut
            clipped_arr -= np.expand_dims(pc_cube, axis=spectrum.spectral_axis_index)

            # and re-clip
            clipped_arr = np.clip(clipped_arr, 0, np.inf)

        self.sonified_cube = CubeListenerData(clipped_arr ** assidx, wlens, duration=0.8,
                                              samplerate=sample_rate, buffsize=buffer_size,
                                              wl_unit=self.sonification_wl_unit,
                                              audfrqmin=audfrqmin, audfrqmax=audfrqmax,
                                              eln=eln, vol=self.volume,
                                              spectral_axis_index=spectrum.spectral_axis_index)

        self.sonified_cube.sonify_cube()
        self.sonified_cube.sigcube = (
                self.sonified_cube.sigcube * pow(whitelight / whitelight.max(),
                                                 ssvidx)).astype('int16')
        self.stream = sd.OutputStream(samplerate=sample_rate, blocksize=buffer_size, device=device,
                                      channels=1, dtype='int16', latency='low',
                                      callback=self.sonified_cube.player_callback)
        self.sonified_cube.cbuff = True

        spatial_inds = [0, 1, 2]
        spatial_inds.remove(spectrum.spectral_axis_index)
        x_size = self.sonified_cube.sigcube.shape[spatial_inds[0]]
        y_size = self.sonified_cube.sigcube.shape[spatial_inds[1]]

        # Create a new entry for the sonified layer in data_lookup. The value is a dictionary
        # containing (x_size * y_size) keys with values being arrays that represent sounds
        if spectrum.spectral_axis_index == 2:
            self.data_lookup[results_label] = {(x, y): self.sonified_cube.sigcube[x, y, :]
                                               for x in range(0, x_size)
                                               for y in range(0, y_size)}
        elif spectrum.spectral_axis_index == 0:
            # This looks wrong but it's because in this case x_size is actually the y axis and vice
            # versa, wasn't sure about the best way to handle the spatial_inds thing above.
            self.data_lookup[results_label] = {(y, x): self.sonified_cube.sigcube[:, x, y]
                                               for x in range(0, x_size)
                                               for y in range(0, y_size)}

        # Create a 2D array with coordinates starting at (0, 0) and going until (x_size, y_size)
        a = np.arange(1, x_size * y_size + 1).reshape((x_size, y_size))

        # Attempt to copy the spatial WCS information from the cube
        if hasattr(spectrum.wcs, 'celestial'):
            wcs = spectrum.wcs.celestial
        elif hasattr(spectrum.wcs, 'to_fits_sip'):
            # GWCS case
            wcs = WCS(spectrum.wcs.to_fits_sip())
        else:
            wcs = None

        sonified_cube = CCDData(a * u.Unit(''), wcs=wcs)
        return sonified_cube

    def update_sonified_cube_with_coord(self, viewer, coord, vollim='buff'):
        # Set newsig to the combined sound array at coord
        if (int(coord[0]), int(coord[1])) not in viewer.combined_sonified_grid:
            return

        # use cached version of combined sonified grid
        compsig = viewer.combined_sonified_grid[int(coord[0]), int(coord[1])]

        # Adjust volume to remove clipping
        if vollim == 'sig':
            # sigmoidal volume limiting
            self.sonified_cube.newsig = (erf(compsig/INT_MAX) * INT_MAX).astype('int16')
        elif vollim == 'clip':
            # hard-clipped volume limiting
            self.sonified_cube.newsig = np.clip(compsig, -INT_MAX, INT_MAX).astype('int16')
        elif vollim == 'buff':
            # renormalise buffer
            sigmax = abs(compsig).max()
            if sigmax > INT_MAX:
                compsig = ((INT_MAX/abs(compsig).max()) * compsig)
            self.sonified_cube.newsig = compsig.astype('int16')
        self.sonified_cube.cbuff = True

    @with_spinner()
    def vue_sonify_cube(self, *args):
        self.sonify_cube()

    def vue_start_stop_stream(self, *args):
        self.stream_active = not self.stream_active

    @observe('spectral_subset_selected')
    def update_wavelength_range(self, event):
        if not hasattr(self, 'spectral_subset'):
            return
        display_unit = self.spectrum_viewer.state.x_display_unit
        # is this spectral selection or the entire spectrum?
        if hasattr(self.spectral_subset.selected_obj, "subregions"):
            wlranges = self.spectral_subset.selected_obj.subregions
        else:
            wlranges = None
        self.sonification_wl_ranges = wlranges
        self.sonification_wl_unit = display_unit

    @observe('volume')
    def update_volume_level(self, event):
        for viewer in self.sonified_viewers:
            viewer.update_volume_level(event['new'])

    @observe('sound_devices_selected')
    def update_sound_device(self, event):
        # This might get called before the plugin is fully initialized
        if not hasattr(self, 'sonified_cube') or not self.sonified_cube:
            return
        if event['new'] != event['old']:
            didx = dict(zip(*self.build_device_lists()))[event['new']]
            # This was moved here from viewers.py now that the stream is handled here.
            self.stop_stream()
            self.stream = sd.OutputStream(samplerate=self.sample_rate, blocksize=self.buffer_size,
                                          device=didx, channels=1, dtype='int16', latency='low',
                                          callback=self.sonified_cube.player_callback)

    def refresh_device_list(self):
        devices, indexes = self.build_device_lists()
        self.sound_device_indexes = dict(zip(devices, indexes))
        self.sound_devices_items = devices
        self.sound_devices_selected = dict(zip(indexes, devices))[sd.default.device[1]]

    def vue_refresh_device_list_in_dropdown(self, *args):
        self.refresh_device_list()

    def build_device_lists(self):
        # dedicated function to build the current *output*
        # device and index lists
        devices = []
        device_indexes = []
        for index, device in enumerate(sd.query_devices()):
            if device['max_output_channels'] > 0 and device['name'] not in devices:
                devices.append(device['name'])
                device_indexes.append(index)
        return devices, device_indexes



import numpy as np

from astropy import units as u
from functools import cached_property
from specutils import Spectrum, SpectrumList

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.loaders.parsers import BaseParser
from jdaviz.core.registries import loader_parser_registry
from jdaviz.utils import standardize_metadata


__all__ = ['SpecutilsSpectrumParser',
           'SpecutilsSpectrumArrayParser',
           'SpecutilsSpectrumListParser']


@loader_parser_registry('specutils.Spectrum')
class SpecutilsSpectrumParser(BaseParser):
    SpecutilsCls = Spectrum

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'specviz', 'specviz2d', 'cubeviz'):
            # NOTE: temporary during deconfig process
            return False
        try:
            self.output
        except Exception:
            return False
        return True

    @cached_property
    def output(self):
        return self.SpecutilsCls.read(self.input)


@loader_parser_registry('specutils.Spectrum(array)')
class SpecutilsSpectrumArrayParser(SpecutilsSpectrumParser):
    @property
    def is_valid(self):
        return (isinstance(self.input, np.ndarray)
                and self.input.ndim in (1, 2, 3)
                and super().is_valid)

    @cached_property
    def output(self):
        arr = self.input

        if not hasattr(arr, 'unit'):
            arr = arr << u.count

        meta = standardize_metadata({})
        # Default to last axis in array for the spectral axis
        msg = "Spectral axis index not specified, assuming last axis."
        self.app.hub.broadcast(SnackbarMessage(msg, sender=self, color="warning"))
        spectral_axis_index = arr.ndim - 1
        return Spectrum(flux=arr, meta=meta, spectral_axis_index=spectral_axis_index)


@loader_parser_registry('specutils.SpectrumList')
class SpecutilsSpectrumListParser(SpecutilsSpectrumParser):
    SpecutilsCls = SpectrumList

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'specviz'):
            # NOTE: temporary during deconfig process
            return False
        return super().is_valid and len(self.output) > 1

    @cached_property
    def output(self):
        return self.SpecutilsCls.read(self.input, flux_col='flux')



import gwcs
import pytest

from astropy.modeling import models
from astropy.nddata import VarianceUncertainty
from astropy.tests.helper import assert_quantity_allclose
import astropy.units as u
from astropy.utils.data import download_file
import numpy as np
from numpy.testing import assert_allclose
from packaging.version import Version
from specreduce import tracing, background, extract
from specutils import Spectrum, SpectralRegion
from glue.core.link_helpers import LinkSameWithUnits
from glue.core.roi import XRangeROI

from jdaviz.core.custom_units_and_equivs import SPEC_PHOTON_FLUX_DENSITY_UNITS
from jdaviz.utils import cached_uri
from jdaviz.core.marks import Lines

GWCS_LT_0_18_1 = Version(gwcs.__version__) < Version('0.18.1')


@pytest.mark.remote_data
@pytest.mark.filterwarnings('ignore')
def test_plugin(specviz2d_helper):
    # TODO: Change back to smaller number (30?) when ITSD is convinced it is them and not us.
    #       Help desk ticket INC0183598, J. Quick.
    fn = download_file('https://stsci.box.com/shared/static/exnkul627fcuhy5akf2gswytud5tazmw.fits',
                       cache=True, timeout=100)

    specviz2d_helper.load_data(spectrum_2d=fn)

    pext = specviz2d_helper.app.get_tray_item_from_name('spectral-extraction-2d')

    # test trace marks - won't be created until after opening the plugin
    sp2dv = specviz2d_helper.app.get_viewer('spectrum-2d-viewer')
    assert len(sp2dv.figure.marks) == 3

    pext.keep_active = True
    assert len(sp2dv.figure.marks) == 12
    assert pext.marks['trace'].marks_list[0].visible is True
    assert len(pext.marks['trace'].marks_list[0].x) > 0

    # create FlatTrace
    pext.trace_type_selected = 'Flat'
    pext.trace_pixel = 28
    trace = pext.export_trace(add_data=True)
    assert isinstance(trace, tracing.FlatTrace)
    assert trace.trace_pos == 28
    trace.trace_pos = 27
    pext.import_trace(trace)
    assert pext.trace_pixel == 27

    # offset existing trace
    pext.trace_trace.selected = 'trace'
    pext.trace_offset = 2
    trace = pext.export_trace(add_data=True)  # overwrite
    assert isinstance(trace, tracing.FlatTrace)

    # create FitTrace
    pext.trace_trace.selected = 'New Trace'
    pext.trace_type_selected = 'Polynomial'
    trace = pext.export_trace(add_data=True)
    assert isinstance(trace, tracing.FitTrace)
    assert trace.guess == 27
    trace = pext.export_trace(trace_pixel=26, add_data=False)
    assert trace.guess == 26
    trace.guess = 28
    pext.import_trace(trace)
    assert pext.trace_pixel == 28

    # offset existing trace
    pext.trace_trace.selected = 'trace'
    pext.trace_offset = 2
    trace = pext.export_trace(add_data=True)  # overwrite
    assert isinstance(trace, tracing.ArrayTrace)

    # TODO: Investigate extra hidden mark from glue-jupyter, see
    # https://github.com/spacetelescope/jdaviz/pull/2478#issuecomment-1731864411
    # 3 new trace objects should have been loaded and plotted in the spectrum-2d-viewer
    # and there are 2 invisible marks from the cross-dispersion profile plugin
    assert len(sp2dv.figure.marks) in [18, 20]

    # interact with background section, check marks
    pext.trace_trace.selected = 'New Trace'
    pext.trace_type_selected = 'Flat'
    pext.bg_separation = 5
    pext.bg_width = 3
    pext.bg_type_selected = 'TwoSided'
    for mark in ['bg1_center', 'bg2_center']:
        assert pext.marks[mark].marks_list[0].visible is True
        assert len(pext.marks[mark].marks_list[0].x) > 0
    bg = pext.export_bg()
    pext.import_bg(bg)
    assert pext.bg_type_selected == 'TwoSided'

    pext.bg_type_selected = 'Manual'
    assert len(pext.marks['bg1_center'].marks_list[0].x) == 0
    assert len(pext.marks['bg2_center'].marks_list[0].x) == 0
    assert len(pext.marks['bg1_lower'].marks_list[0].x) > 0

    pext.bg_type_selected = 'OneSided'
    # only bg1 is populated for OneSided
    assert len(pext.marks['bg1_center'].marks_list[0].x) > 0
    assert len(pext.marks['bg2_center'].marks_list[0].x) == 0

    # create background image
    pext.bg_separation = 4
    bg = pext.export_bg()
    assert isinstance(bg, background.Background)
    assert len(bg.traces) == 1
    assert bg.traces[0].trace[0] == 28 + 4
    assert bg.width == 3
    bg = pext.export_bg(bg_width=3.3)
    assert bg.width == 3.3
    bg.width = 4
    pext.import_bg(bg)
    assert pext.bg_width == 4
    bg_img = pext.export_bg_img()
    assert isinstance(bg_img, Spectrum)
    bg_spec = pext.export_bg_spectrum()
    assert isinstance(bg_spec, Spectrum)
    bg_sub = pext.export_bg_sub()
    assert isinstance(bg_sub, Spectrum)

    # interact with extraction section, check marks
    pext.ext_width = 1
    for mark in ['bg1_center', 'bg2_center']:
        assert pext.marks[mark].marks_list[0].visible is False
    for mark in ['ext_lower', 'ext_upper']:
        assert pext.marks[mark].marks_list[0].visible is True
        assert len(pext.marks[mark].marks_list[0].x) > 0

    # create subtracted spectrum
    ext = pext.export_extract(ext_width=3)
    assert isinstance(ext, extract.BoxcarExtract)
    assert ext.width == 3
    ext.width = 2
    pext.import_extract(ext)
    assert pext.ext_width == 2
    sp_ext = pext.export_extract_spectrum()
    assert isinstance(sp_ext, Spectrum)

    pext.ext_type_selected = 'Horne'
    sp_ext = pext.export_extract_spectrum()
    assert isinstance(sp_ext, Spectrum)

    # test API calls
    for step in ['trace', 'bg', 'ext']:
        pext.update_marks(step)

    # test exception handling
    pext.trace_type = 'Polynomial'
    pext.bg_type_selected = 'TwoSided'
    pext.bg_separation = 1
    pext.bg_width = 5
    assert len(pext.ext_specreduce_err) > 0
    pext.bg_results_label = 'should not be created'
    pext.vue_create_bg_img()
    assert 'should not be created' not in [d.label for d in specviz2d_helper.app.data_collection]

    with pytest.raises(ValueError):
        pext.export_extract(invalid_kwarg=5)

    # test importing traces
    img = specviz2d_helper.get_data('Spectrum 2D')
    flat_trace = tracing.FlatTrace(img, trace_pos=25)
    fit_trace = tracing.FitTrace(img)

    for imported_trace in [flat_trace, fit_trace]:
        pext.import_trace(imported_trace)
        exported_trace = pext.export_trace(add_data=False)
        assert isinstance(exported_trace, type(imported_trace))

    array_trace = tracing.ArrayTrace(img, np.arange(len(img.spectral_axis)))
    specviz2d_helper.load(array_trace, data_label='array_trace')
    pext.trace_trace.selected = 'array_trace'
    exported_trace = pext.export_trace(add_data=False)
    assert isinstance(exported_trace, tracing.ArrayTrace)


@pytest.mark.remote_data
@pytest.mark.filterwarnings('ignore')
def test_user_api(specviz2d_helper):
    fn = download_file('https://stsci.box.com/shared/static/exnkul627fcuhy5akf2gswytud5tazmw.fits', cache=True)  # noqa

    specviz2d_helper.load_data(spectrum_2d=fn)

    pext = specviz2d_helper.plugins['2D Spectral Extraction']
    pext.keep_active = True

    # test that setting a string to an AddResults object redirects to the label
    pext.bg_sub_add_results = 'override label'
    assert pext.bg_sub_add_results.label == 'override label'
    pext.bg_sub_add_results.label = 'override label 2'
    assert "override label 2" in pext.bg_sub_add_results._obj.__repr__()
    assert "override label 2" in pext.bg_sub_add_results._obj.auto_label.__repr__()

    pext.export_bg_sub(add_data=True)
    assert 'override label 2' in pext.ext_dataset.choices

    # test setting auto label
    pext.bg_sub_add_results.auto = True


@pytest.mark.remote_data
@pytest.mark.skipif(GWCS_LT_0_18_1, reason='Needs GWCS 0.18.1 or later')
@pytest.mark.filterwarnings("ignore::astropy.wcs.wcs.FITSFixedWarning")
def test_background_extraction_and_display(specviz2d_helper):
    uri = 'mast:jwst/product/jw01538-o161_t002-s000000001_nirspec_f290lp-g395h-s1600a1_s2d.fits'
    specviz2d_helper.load_data(spectrum_2d=cached_uri(uri), cache=True)
    pext = specviz2d_helper.app.get_tray_item_from_name('spectral-extraction-2d')

    # check that the background extraction method and parameters are as expected
    assert pext.bg_type_selected == 'TwoSided'
    np.testing.assert_allclose(pext.bg_separation, 6)

    # test extracting background and background subtracted images and adding
    # them to the viewer
    pext.export_bg_sub(add_data=True)
    assert specviz2d_helper.app.data_collection[2].label == 'background-subtracted'

    pext.export_bg_img(add_data=True)
    assert specviz2d_helper.app.data_collection[3].label == 'background'


@pytest.mark.filterwarnings('ignore')
def test_horne_extract_self_profile(specviz2d_helper):

    spec2d = np.zeros((40, 100))
    spec2dvar = np.ones((40, 100))

    for ii in range(spec2d.shape[1]):
        mgaus = models.Gaussian1D(amplitude=10,
                                  mean=(9.+(20/spec2d.shape[1])*ii),
                                  stddev=2)
        rg = np.arange(0, spec2d.shape[0], 1)
        gaus = mgaus(rg)
        spec2d[:, ii] = gaus

    wave = np.arange(0, spec2d.shape[1], 1)
    objectspec = Spectrum(spectral_axis=wave*u.m,
                          flux=spec2d*u.Jy,
                          uncertainty=VarianceUncertainty(spec2dvar*u.Jy*u.Jy))

    specviz2d_helper.load(objectspec)
    pext = specviz2d_helper.plugins['2D Spectral Extraction']._obj

    trace_fit = tracing.FitTrace(objectspec,
                                 trace_model=models.Polynomial1D(degree=1),
                                 window=13, peak_method='gaussian', guess=20)
    pext.import_trace(trace_fit)

    pext.ext_type.selected = "Horne"
    pext.horne_ext_profile.selected = "Self (interpolated)"

    # check that correct defaults are set
    assert pext.self_prof_n_bins == 10
    assert pext.self_prof_interp_degree_x == 1
    assert pext.self_prof_interp_degree_y == 1

    sp_ext = pext.export_extract_spectrum()

    bg_sub = pext.export_bg_sub()

    extract_horne_interp = extract.HorneExtract(bg_sub, trace_fit,
                                                spatial_profile='interpolated_profile')

    assert_quantity_allclose(extract_horne_interp.spectrum.flux, sp_ext.flux)

    # now try changing from defaults
    pext.self_prof_n_bins = 5
    pext.self_prof_interp_degree_x = 2
    pext.self_prof_interp_degree_y = 2

    sp_ext = pext.export_extract_spectrum()
    bg_sub = pext.export_bg_sub()

    extract_horne_interp = extract.HorneExtract(bg_sub, trace_fit,
                                                spatial_profile={'name': 'interpolated_profile',
                                                                 'n_bins_interpolated_profile': 5,
                                                                 'interp_degree': (2, 2)})

    assert_quantity_allclose(extract_horne_interp.spectrum.flux, sp_ext.flux)

    # test that correct errors are raised
    pext.self_prof_n_bins = 0
    with pytest.raises(ValueError, match='must be greater than 0'):
        sp_ext = pext.export_extract_spectrum()

    pext.self_prof_n_bins = 1
    pext.self_prof_interp_degree_x = 0
    with pytest.raises(ValueError, match='must be greater than 0'):
        sp_ext = pext.export_extract_spectrum()

    pext.self_prof_interp_degree_x = 1
    pext.self_prof_interp_degree_y = 0
    with pytest.raises(ValueError, match='`self_prof_interp_degree_y` must be greater than 0.'):
        sp_ext = pext.export_extract_spectrum()


def test_spectral_extraction_flux_unit_conversions(specviz2d_helper, mos_spectrum2d):
    specviz2d_helper.load(mos_spectrum2d)

    uc = specviz2d_helper.plugins["Unit Conversion"]
    pext = specviz2d_helper.plugins['2D Spectral Extraction']

    for new_flux_unit in SPEC_PHOTON_FLUX_DENSITY_UNITS:

        # iterate through flux units verifying that selected object/spectrum is obtained using
        # display units
        uc.flux_unit.selected = new_flux_unit

        exported_bg = pext.export_bg()
        assert exported_bg.image._unit == specviz2d_helper.app._get_display_unit('flux')

        exported_bg_img = pext.export_bg_img()
        assert exported_bg_img._unit == specviz2d_helper.app._get_display_unit('flux')

        exported_bg_sub = pext.export_bg_sub()
        assert exported_bg_sub._unit == specviz2d_helper.app._get_display_unit('flux')

        exported_extract_spectrum = pext.export_extract_spectrum()
        assert exported_extract_spectrum._unit == specviz2d_helper.app._get_display_unit('flux')

        exported_extract = pext.export_extract()
        assert exported_extract.image._unit == specviz2d_helper.app._get_display_unit('flux')


@pytest.mark.filterwarnings('ignore')
def test_spectral_extraction_preview(deconfigged_helper, spectrum2d):
    # Multiple 1D viewers is a deconfigged thing so we stick with using
    # that helper throughout.
    deconfigged_helper.load(spectrum2d, format='2D Spectrum')
    # Assume auto extract is on
    default_extraction_label = deconfigged_helper.app.data_collection.labels[1]
    custom_extraction_label = 'custom-extraction'

    # Create a custom extraction and add to a existing viewer
    spext = deconfigged_helper.plugins['2D Spectral Extraction']
    spext.trace_pixel = 2  # to distinguish from auto-extract
    spext.ext_add_results.label = custom_extraction_label
    spext.export_extract_spectrum(add_data=True)  # added as layer to "1D Spectrum" viewer

    default_extraction_data = deconfigged_helper.get_data(default_extraction_label)
    custom_extraction_data = deconfigged_helper.get_data(custom_extraction_label)

    assert np.any(~np.isnan(default_extraction_data.flux))
    assert np.any(~np.isnan(custom_extraction_data.flux))
    # Check that the loaded spectra differ from one another, i.e. that
    # the second spectra is being pulled in correctly.
    assert not np.array_equal(default_extraction_data.flux, custom_extraction_data.flux)

    # Check that both are in the default viewer
    default_viewer = deconfigged_helper.viewers['1D Spectrum']
    default_dm = default_viewer.data_menu
    assert default_extraction_label in default_dm.layer.choices
    assert custom_extraction_label in default_dm.layer.choices

    # Remove layer from default viewer
    default_dm.layer = [custom_extraction_label]
    default_dm.remove_from_viewer()

    # Check that only the default extraction remains in the default viewer
    default_dm = default_viewer.data_menu
    assert default_extraction_label in default_dm.layer.choices
    assert custom_extraction_label not in default_dm.layer.choices

    # and add to a new viewer
    nvc = deconfigged_helper.new_viewers['1D Spectrum']
    nvc.dataset = custom_extraction_label
    nvc.viewer_label = 'manual-viewer'
    new_viewer = nvc()

    # Iterate through both viewers to perform the existence and preview checks.
    # Adds the custom extraction *back* to the default viewer to check
    # that the preview works there as well even after all the moving around.
    for i, viewer in enumerate([new_viewer, default_viewer]):
        dm = viewer.data_menu
        if i == 0:
            # Check that only the custom extraction is in the new viewer
            assert default_extraction_label not in dm.layer.choices
        else:
            # re-add to old viewer and confirm that the preview appears there too
            default_dm.add_data(custom_extraction_label)
            assert default_extraction_label in dm.layer.choices

        assert custom_extraction_label in dm.layer.choices

        # check that the preview lines appear/disappear as expected
        # (as modified from 'test_ramp_extraction.py')
        for interactive_extract in [True, False]:
            with spext.as_active():
                spext.interactive_extract = interactive_extract
                assert len([
                    mark for mark in viewer._obj.glue_viewer.custom_marks
                    if mark.visible and isinstance(mark, Lines) and
                       len(mark.x) == len(spectrum2d.spectral_axis)]) == int(interactive_extract)


class TestTwo2dSpectra:

    def load_2d_spectrum(self, helper, spec2d, spec2d_label_idx=0, spec2d_ext_label_idx=1):
        # Allow this to use the default label
        helper.load(spec2d, format='2D Spectrum')
        dc_labels = helper.app.data_collection.labels
        self.spec2d_label = dc_labels[spec2d_label_idx]
        self.spec2d_ext_label = dc_labels[spec2d_ext_label_idx]

    def setup_another_2d_spectrum(self, spec2d):
        # Specific data labels
        another_spec2d_label = 'Another 2D Spectrum'
        another_spec2d_ext_label = 'Another 2D Spectral Extraction'

        # Make the data different enough that extracted spectra should differ
        another_spec2d = Spectrum(flux=spec2d.flux * 2, spectral_axis=spec2d.spectral_axis)

        self.another_spec2d = another_spec2d
        self.another_spec2d_label = another_spec2d_label
        self.another_spec2d_ext_label = another_spec2d_ext_label

    def load_another_2d_spectrum(self, helper, spec2d):
        self.setup_another_2d_spectrum(spec2d)
        helper.load(self.another_spec2d,
                    format='2D Spectrum',
                    data_label=self.another_spec2d_label,
                    ext_data_label=self.another_spec2d_ext_label)

    @pytest.mark.parametrize(('method', 'helper'),
                             [('load', 'deconfigged_helper'),
                              ('specviz2d', 'specviz2d_helper'),
                              ('specviz2d_alternate_order', 'specviz2d_helper'),
                              ('loader_infrastructure', 'deconfigged_helper'),
                              ('loader_infrastructure_alternate_order', 'deconfigged_helper')])
    def test_labels_and_spectral_extraction_flux_difference(self, method, helper, request,
                                                            spectrum2d):
        helper = request.getfixturevalue(helper)
        if method in ('load', 'specviz2d'):
            self.load_2d_spectrum(helper, spectrum2d)
            self.load_another_2d_spectrum(helper, spectrum2d)

        elif method == 'specviz2d_alternate_order':
            self.load_another_2d_spectrum(helper, spectrum2d)
            self.load_2d_spectrum(helper, spectrum2d, spec2d_label_idx=2, spec2d_ext_label_idx=3)

        elif method == 'loader_infrastructure':
            # Allow this to use the default label
            ldr = helper.loaders['object']
            ldr.object = spectrum2d
            ldr.format = '2D Spectrum'
            ldr.importer.auto_extract = True
            ldr.load()
            self.spec2d_label, self.spec2d_ext_label = helper.app.data_collection.labels[:2]

            self.setup_another_2d_spectrum(spectrum2d)
            ldr.object = self.another_spec2d
            ldr.format = '2D Spectrum'
            ldr.importer.auto_extract = True
            ldr.importer.data_label = self.another_spec2d_label
            ldr.importer.ext_data_label = self.another_spec2d_ext_label
            ldr.load()

        elif method == 'loader_infrastructure_alternate_order':
            self.setup_another_2d_spectrum(spectrum2d)
            # Swapping the load order changes the parent and
            # the spectral extraction that is duplicated
            ldr = helper.loaders['object']
            ldr.object = self.another_spec2d
            ldr.format = '2D Spectrum'
            ldr.importer.auto_extract = True
            ldr.importer.data_label = self.another_spec2d_label
            ldr.importer.ext_data_label = self.another_spec2d_ext_label
            ldr.load()

            # Allow this to use the default label
            ldr.object = spectrum2d
            ldr.format = '2D Spectrum'
            ldr.importer.auto_extract = True
            ldr.load()
            self.spec2d_label, self.spec2d_ext_label = helper.app.data_collection.labels[-2:]

        else:
            raise NotImplementedError(f"Method {method} not implemented.")

        expected_labels = [self.spec2d_label, self.spec2d_ext_label,
                           self.another_spec2d_label, self.another_spec2d_ext_label]

        dc = helper.app.data_collection

        assert self.spec2d_label in dc.labels
        assert self.spec2d_ext_label in dc.labels
        spec2d = helper.get_data(self.spec2d_label)
        extracted_spec2d = helper.get_data(self.spec2d_ext_label)
        # Check for any non-NaN data, if all NaNs, something went wrong
        assert np.any(~np.isnan(spec2d.flux))
        assert np.any(~np.isnan(extracted_spec2d.flux))

        assert self.another_spec2d_label in dc.labels
        assert self.another_spec2d_ext_label in dc.labels
        another_spec2d = helper.get_data(self.another_spec2d_label)
        extracted_another_spec2d = helper.get_data(self.another_spec2d_ext_label)
        # Check for any non-NaN data, if all NaNs, something went wrong
        assert np.any(~np.isnan(another_spec2d.flux))
        assert np.any(~np.isnan(extracted_another_spec2d.flux))

        # Check that the loaded spectra differ from one another, i.e. that
        # the second spectra is being pulled in correctly.
        assert not np.array_equal(spec2d.flux, another_spec2d.flux), \
            'Loaded spectra should differ!'
        assert not np.array_equal(extracted_spec2d.flux, extracted_another_spec2d.flux), \
            'Extracted spectra should differ!'

        # Check linking, e.g.
        # 2D Spectrum (auto-ext) <=> 2D Spectrum [spectral axis]
        # 2D Spectrum (auto-ext) <=> 2D Spectrum [spectral flux density]
        # Another 2D Spectrum <=> 2D Spectrum [spectral axis]
        # Another 2D Spectrum <=> 2D Spectrum [spectral flux density]
        # Another 2D Spectral Extraction <=> 2D Spectrum [spectral axis]
        # Another 2D Spectral Extraction <=> 2D Spectrum
        assert len(dc.external_links) == 9
        for link in dc.external_links:
            # Check that linking is correct by confirming that both
            # are in `expected_labels`
            assert (link.data1.label in expected_labels) and (link.data2.label in expected_labels)
            assert isinstance(link, LinkSameWithUnits)

    def test_subsets_and_viewer_things(self, deconfigged_helper, spectrum2d):
        # Allow this to use the default label
        self.load_2d_spectrum(deconfigged_helper, spectrum2d)
        self.load_another_2d_spectrum(deconfigged_helper, spectrum2d)

        # Test marks/subsets between the two layers
        viewer_2d = deconfigged_helper.app.get_viewer('2D Spectrum')
        viewer_1d = deconfigged_helper.app.get_viewer('1D Spectrum')

        # We'll be panning along x so keep x_min, x_max generic but specify y_min_1d, y_max_1d
        x_min_2d, x_max_2d = viewer_2d.get_limits()[:2]
        midway = (x_min_2d + x_max_2d) / 2

        # Confirm that no subsets are present to start with
        assert not any('subset' in str(layer).lower() for layer in viewer_1d.layers)
        assert not any('subset' in str(layer).lower() for layer in viewer_2d.layers)

        # Following the procedure from ``test_subsets.py``
        # create subset in 2d viewer, want data in 1d viewer
        viewer_2d.apply_roi(XRangeROI(x_min_2d + 0.5 * midway, x_max_2d - 0.5 * midway))
        # Confirm that subsets are present in both viewers
        subset_label = 'Subset 1'
        assert any(subset_label in str(layer) for layer in viewer_1d.layers)
        assert any(subset_label in str(layer) for layer in viewer_2d.layers)

        subset_drawn_2d = viewer_1d.native_marks[-1]
        # get x and y components to compute subset mask
        y1 = subset_drawn_2d.y
        x1 = subset_drawn_2d.x

        subset_highlighted_region1 = x1[np.isfinite(y1)]
        min_value_subset = np.min(subset_highlighted_region1)
        max_value_subset = np.max(subset_highlighted_region1)

        expected_min = 2
        expected_max = 7
        atol = 1e-2
        assert_allclose(min_value_subset, expected_min, atol=atol)
        assert_allclose(max_value_subset, expected_max, atol=atol)

        # now create a subset in the spectrum-viewer, and determine if
        # subset is linked correctly in spectrum2d-viewer
        x_min_reg = x_min_2d + midway
        x_max_reg = x_max_2d - midway
        spec_reg = SpectralRegion(x_min_reg * u.um, x_max_reg * u.um)
        st = deconfigged_helper.plugins['Subset Tools']
        st.import_region(spec_reg, edit_subset=subset_label)

        # Check again
        assert any(subset_label in str(layer) for layer in viewer_1d.layers)
        assert any(subset_label in str(layer) for layer in viewer_2d.layers)

        mask = viewer_2d._get_layer(subset_label)._get_image()

        x_coords = np.nonzero(mask)[1]
        min_value_subset = x_coords.min()
        max_value_subset = x_coords.max()

        assert_allclose(min_value_subset, x_min_reg, atol=atol)
        assert_allclose(max_value_subset, x_max_reg, atol=atol)

        # Now trying through the viewer ROI
        roi = XRangeROI(0, 2)
        viewer_2d.toolbar.active_tool = viewer_2d.toolbar.tools['bqplot:xrange']
        viewer_2d.toolbar.active_tool.activate()
        viewer_2d.toolbar.active_tool.update_from_roi(roi)
        viewer_2d.toolbar.active_tool.update_selection()

        # Check that the subset appears in both viewers
        subset_label = 'Subset 2'
        assert any(subset_label in str(layer) for layer in viewer_1d.layers)
        assert any(subset_label in str(layer) for layer in viewer_2d.layers)

        # Not 100% sure why there is a difference of 1 in the number of marks
        # but leaving this here for posterity.
        assert len(viewer_2d.native_marks) == 7
        assert len(viewer_1d.native_marks) == 6

        # Checking that panning one viewer pans the other
        viewer_2d.toolbar.active_tool = viewer_2d.toolbar.tools['jdaviz:panzoom_matchx']

        # Simulate a pan by changing the x-axis limits
        x_min_1d, x_max_1d, y_min_1d, y_max_1d = viewer_1d.get_limits()
        dx = 1
        # 2D viewer pan
        viewer_2d.set_limits(x_min=x_min_2d+dx, x_max=x_max_2d+dx)
        # Double check that the 2D viewer limits were actually changed
        assert_allclose(viewer_2d.get_limits()[:2], (x_min_2d+dx, x_max_2d+dx))

        # Notify the active tool that the limits have changed (simulate pan event)
        viewer_2d.toolbar.active_tool.on_limits_change()

        # Check that the 1D viewer updated its x-axis limits to match the 2D viewer
        expected = (x_min_1d + dx, x_max_1d + dx, y_min_1d, y_max_1d)
        assert_allclose(viewer_1d.get_limits(), expected)



from .importer import *  # noqa
from .spectrum_common import *  # noqa
from .spectrum1d import *  # noqa
from .spectrum2d import *  # noqa
from .spectrum3d import *  # noqa
from .spectrum_list import *  # noqa
from .ramp import *  # noqa
from .ramp_integration import *  # noqa
from .image import *  # noqa
from .subset import *  # noqa
from .trace import *  # noqa
from .catalog import *  # noqa
from .footprint import *  # noqa


from .viewer_creators import BaseViewerCreator  # noqa
from .spectrum1d.spectrum1d import *  # noqa
from .spectrum2d.spectrum2d import *  # noqa
from .spectrum3d.spectrum3d import *  # noqa
from .image.image import *  # noqa
from .histogram.histogram import *  # noqa
from .scatter.scatter import *  # noqa
from .table.table import *  # noqa
from .ramp3d.ramp3d import *  # noqa
from .ramp_integration.ramp_integration import *  # noqa



from specreduce.tracing import Trace


from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.loaders.importers import BaseImporterToDataCollection


@loader_importer_registry('Trace')
class TraceImporter(BaseImporterToDataCollection):
    template_file = __file__, "../to_dc_with_label.vue"

    @staticmethod
    def _get_supported_viewers():
        return [{'label': '2D Spectrum', 'reference': 'spectrum-2d-viewer'}]

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'specviz2d'):
            # NOTE: temporary during deconfig process
            return False
        return (isinstance(self.input, Trace)
                and 'Spectral Extraction' in self.app._jdaviz_helper.plugins)



import warnings

import asdf
import numpy as np
from astropy import units as u
from astropy.io import fits
from astropy.nddata import NDData, CCDData
from astropy.utils.exceptions import AstropyWarning
from astropy.wcs import WCS
from glue.core.data import Component, Data
from gwcs import WCS as GWCS
from stdatamodels import asdf_in_fits
from traitlets import Bool, List, Any, Unicode, observe

from jdaviz.core.events import SnackbarMessage
from jdaviz.core.template_mixin import (SelectFileExtensionComponent,
                                        DatasetSelect,
                                        SelectPluginComponent)

from jdaviz.core.loaders.importers import BaseImporterToDataCollection
from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.user_api import ImporterUserApi

from jdaviz.utils import (
    PRIHDR_KEY, standardize_metadata, standardize_roman_metadata,
    _try_gwcs_to_fits_sip, create_data_hash, RA_COMPS, DEC_COMPS
)

try:
    from roman_datamodels import datamodels as rdd
except ImportError:
    HAS_ROMAN_DATAMODELS = False
else:
    HAS_ROMAN_DATAMODELS = True


__all__ = ['ImageImporter', '_spatial_assign_component_type']


def _spatial_assign_component_type(comp_id, comp, units, physical_type):
    if comp_id.startswith('Pixel Axis'):
        physical_type = 'pixel'
        return f'{comp_id[-2]}:pixel'

    if comp_id.lower() in RA_COMPS and physical_type == 'angle':
        return f'RA:{physical_type}'
    elif comp_id.lower() in DEC_COMPS and physical_type == 'angle':
        return f'DEC:{physical_type}'

    return physical_type


@loader_importer_registry('Image')
class ImageImporter(BaseImporterToDataCollection):
    template_file = __file__, "./image.vue"

    # HDUList-specific options
    input_has_extensions = Bool(False).tag(sync=True)
    extension_items = List().tag(sync=True)
    extension_selected = Any().tag(sync=True)
    extension_multiselect = Bool(True).tag(sync=True)

    # Data-association
    parent_items = List().tag(sync=True)
    parent_selected = Any().tag(sync=True)

    # Use FITS approximation instead of original image GWCS
    gwcs_to_fits_sip = Bool(False).tag(sync=True)

    # Alignment options
    align_by_items = List().tag(sync=True)
    align_by_selected = Unicode().tag(sync=True)

    # user-settable option to treat the data_label as prefix and append the extension later
    data_label_as_prefix = Bool(False).tag(sync=True)
    # whether the current data_label should be treated as a prefix
    # either based on user-setting above or current extension selection
    data_label_is_prefix = Bool(False).tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.parent = DatasetSelect(self, 'parent_items', 'parent_selected',
                                    multiselect=None, manual_options=['Auto'])
        self.parent.add_filter('is_image', 'not_from_plugin')
        self.parent.selected = 'Auto'

        self.align_by = SelectPluginComponent(self,
                                              items='align_by_items',
                                              selected='align_by_selected',
                                              manual_options=['Pixels', 'WCS'])
        align_plg = self.app._jdaviz_helper.plugins.get('Orientation', None)
        if align_plg is not None:
            self.align_by.selected = align_plg.align_by.selected
        else:
            self.align_by.selected = 'Pixels'

        input = self.input
        if isinstance(input, fits.hdu.image.ImageHDU):
            input = fits.HDUList([input])

        input_is_roman = (HAS_ROMAN_DATAMODELS and
                          isinstance(input, (rdd.ImageModel, rdd.DataModel)))
        self.input_has_extensions = (isinstance(input, fits.HDUList) or
                                     input_is_roman)
        if self.input_has_extensions:
            if isinstance(input, fits.HDUList):
                filters = [_validate_fits_image2d]
                ext_options = [{'label': f"{index}: [{hdu.name},{hdu.ver}]",
                                'name': hdu.name,
                                'ver': hdu.ver,
                                'name_ver': f"{hdu.name},{hdu.ver}",
                                'index': index,
                                'data_hash': create_data_hash(hdu),
                                'obj': hdu}
                               for index, hdu in enumerate(input)]
            elif input_is_roman:
                filters = [_validate_roman_ext]
                ext_options = [{'label': f"{index}: {key}",
                                'name': key,
                                'ver': None,
                                'name_ver': key,
                                'index': index,
                                'data_hash': create_data_hash(value),
                                'obj': value}
                               for index, (key, value) in enumerate(input.items())]
            else:
                raise NotImplementedError()

            self.extension = SelectFileExtensionComponent(self,
                                                          items='extension_items',
                                                          selected='extension_selected',
                                                          multiselect='extension_multiselect',
                                                          manual_options=ext_options,
                                                          filters=filters)

            # changing selected extension will call _set_default_data_label
            self.extension.selected = [self.extension.choices[0]]
            self.data_hashes = self.extension.data_hashes
            self.hash_map_to_label = dict(zip(self.extension.data_hashes, self.extension.labels))
        else:
            self._set_default_data_label()

    def _get_supported_viewers(self):
        if self.config == 'rampviz':
            return [{'label': 'level-2', 'reference': 'imviz-image-viewer'}]
        return [{'label': 'Image', 'reference': 'imviz-image-viewer'}]

    @property
    def user_api(self):
        expose = ['parent', 'data_label_as_prefix', 'gwcs_to_fits_sip', 'align_by']
        if self.input_has_extensions:
            expose += ['extension']
        return ImporterUserApi(self, expose)

    @property
    def is_valid(self):
        if self.app.config not in ('deconfigged', 'imviz', 'mastviz', 'cubeviz', 'rampviz'):
            # NOTE: temporary during deconfig process
            return False
        # flat image, not a cube
        # isinstance NDData
        if self.input_has_extensions and not len(self.extension.choices):
            return False

        if isinstance(self.input, (fits.HDUList, fits.hdu.image.ImageHDU,
                                   NDData, np.ndarray, Data)):
            supported_input_type = True
        elif isinstance(self.input, (asdf.AsdfFile)) or \
                (HAS_ROMAN_DATAMODELS and isinstance(self.input, (rdd.DataModel, rdd.ImageModel))):
            supported_input_type = True
        else:
            supported_input_type = False

        if not supported_input_type:
            return False

        try:
            _ = self.output
        except Exception:  # noqa
            return False
        else:
            return True

    def _glue_data_wcs_to_fits(self, glue_data):
        """
        Re-set data.coords to a SIP approximation of the GWCS, if
        gwcs_to_fits_sip is True and data.coords is a GWCS. If these conditions
        are met but this approximation is not possible, a warning will be
        emitted and the original GWCS will be used.
        """
        if isinstance(glue_data.coords, GWCS):
            glue_data.coords = _try_gwcs_to_fits_sip(glue_data.coords)
        return glue_data

    def _get_label_with_extension(self, prefix, ext=None, ver=None):
        full_ext = ",".join([str(e) for e in (ext, ver) if e is not None])
        return f"{prefix}[{full_ext}]" if len(full_ext) else prefix

    @observe('extension_selected', 'data_label_as_prefix')
    def _set_default_data_label(self, *args):
        if self.default_data_label_from_resolver:
            prefix = self.default_data_label_from_resolver
        else:
            prefix = "Image"

        if self.input_has_extensions and hasattr(self, 'extension'):
            if self.extension.selected_name is None:
                return
            if len(self.extension.selected_name) == 1 and not self.data_label_as_prefix:
                # selected_obj may be an ndarray object if input is a roman data model
                ver = getattr(self.extension.selected_obj[0], 'ver', None)
                # only a single extension selected
                self.data_label_default = self._get_label_with_extension(prefix,
                                                                         ext=self.extension.selected_name[0],  # noqa
                                                                         ver=ver)  # noqa
                self.data_label_is_prefix = False
            else:
                # multiple extensions selected,
                # only show the prefix and append the extension later during import
                self.data_label_default = prefix
                self.data_label_is_prefix = True
        elif (self.data_label_as_prefix or
              (isinstance(self.input, NDData) and
               getattr(self.input, 'meta', {}).get('plugin', None) is None)):
            # will append with [DATA]/[UNCERTAINTY]/[MASK] later
            # TODO: allow user to select extensions and include in same logic as HDUList
            self.data_label_default = prefix
            self.data_label_is_prefix = True
        else:
            self.data_label_default = prefix
            self.data_label_is_prefix = False

    @property
    def output(self):
        # NOTE: this should ALWAYS return a list of objects able to be imported into DataCollection
        # NDData or ndarray
        input = self.input

        # ImageHDU - treat as HDUList with single extension
        if isinstance(input, fits.hdu.image.ImageHDU):
            input = fits.HDUList([self.input])

        if isinstance(input, Data):
            # Already a Glue object
            data = [input,]
        elif isinstance(input, NDData):
            data = _nddata_to_glue_data(input)  # list of Data
        elif isinstance(input, np.ndarray):
            data = [_ndarray_to_glue_data(input)]
        # asdf
        elif (isinstance(input, asdf.AsdfFile)):
            data = [_roman_asdf_2d_to_glue_data(input, ext='data')]
        # roman data models
        elif HAS_ROMAN_DATAMODELS and isinstance(input, (rdd.DataModel, rdd.ImageModel)):
            data = [_roman_asdf_2d_to_glue_data(input, ext=ext)
                    for ext in self.extension.selected_name]
        # JWST ASDF-in-FITS
        elif isinstance(input, fits.HDUList) and 'ASDF' in input:
            data = [_jwst2data(hdu, input) for hdu in self.extension.selected_obj]
        # FITS
        else:
            data = [_hdu2data(hdu, input) for hdu in self.extension.selected_obj]

        ext_names = self.extension.selected_name if self.input_has_extensions else [None] * len(data)  # noqa
        for d, ext_name in zip(data, ext_names):
            if d is None:
                continue
            d.meta['_extname'] = ext_name
            for component_id in d.main_components:
                if component_id.label.lower().startswith("dq"):
                    # for DQ components, map zeros to nans
                    # so that they are not displayed in the DQ colormap
                    cid = d.get_component(component_id)
                    data_arr = np.float32(cid.data)
                    data_arr[data_arr == 0] = np.nan
                    d.update_components({cid: data_arr})

        return data

    def __call__(self):

        base_data_label = self.data_label_value
        # self.output is always a list of Data objects
        outputs = self.output

        if self.input_has_extensions:
            ext_items = self.extension.selected_item_list
        elif isinstance(self.input, NDData):
            ext_items = [{'name': name} for name in ('DATA', 'MASK', 'UNCERTAINTY')]  # noqa must match order in _nddata_to_glue_data
        else:
            ext_items = [{}] * len(outputs)

        parent_selected = self.parent.selected
        for output, ext_item in zip(outputs, ext_items):
            if output is None:
                # needed for NDData where one of the "extensions" might
                # not be present.  Remove this once users can select
                # which to import.
                continue

            # Determine data label
            if self.data_label_is_prefix:
                # If data_label is a prefix, we need to append the extension
                # to the data label.
                data_label = self._get_label_with_extension(base_data_label,
                                                            ext_item.get('name'),
                                                            ver=ext_item.get('ver', None))
            else:
                # If data_label is not a prefix, we use it as is.
                data_label = base_data_label

            # Determine parent
            if (parent_selected == 'Auto' and
                    len(ext_items) > 1 and
                    getattr(self.input, 'meta', {}).get('plugin', None) is None):
                # If parent is set to 'Auto', use SCI/DATA extension
                # as parent of any other selected extensions with same ver (if applicable)
                for other_ext_item in ext_items:
                    if (other_ext_item.get('name') in ('SCI', 'sci', 'DATA', 'data')
                            and other_ext_item.get('ver', None) == ext_item.get('ver', None)):
                        parent_ext_item = other_ext_item
                        break
                else:
                    # No SCI/DATA extension found, so default to no parent
                    parent_ext_item = None
                    parent_data_label = None
                if parent_ext_item is not None:
                    # assume self.data_label_is_prefix is True
                    parent_data_label = self._get_label_with_extension(base_data_label,
                                                                       parent_ext_item.get('name'),
                                                                       ver=parent_ext_item.get('ver', None))  # noqa
            elif parent_selected == 'Auto':
                parent_data_label = None
            else:
                parent_data_label = parent_selected

            if self.gwcs_to_fits_sip:
                output = self._glue_data_wcs_to_fits(output)

            self.add_to_data_collection(output, data_label, data_hash=ext_item.get('data_hash'),
                                        parent=parent_data_label if parent_data_label != data_label else None,  # noqa
                                        cls=CCDData)

        align_plg = self.app._jdaviz_helper.plugins.get('Orientation', None)
        if align_plg is not None:
            align_plg.align_by.selected = self.align_by.selected
        else:
            msg = f"could not change align_by to {self.align_by.selected}"
            self.app.hub.broadcast(SnackbarMessage(msg, sender=self, color='warning'))

    def assign_component_type(self, comp_id, comp, units, physical_type):
        return _spatial_assign_component_type(comp_id, comp, units, physical_type)


def _validate_fits_image2d(item):
    hdu = item.get('obj')
    return hdu.data is not None and hdu.is_image and hdu.data.ndim == 2


def _validate_roman_ext(item):
    name = item.get('name')
    return name in ['data', 'dq', 'err', 'var_poisson', 'var_rnoise']


def _hdu2data(hdu, hdulist, include_wcs=True):
    if 'BUNIT' in hdu.header:
        bunit = _validate_bunit(hdu.header['BUNIT'], raise_error=False)
    else:
        bunit = ''

    comp_label = f'{hdu.name.upper()},{hdu.ver}'

    data = Data()
    if hdulist is not None and hdu.name != 'PRIMARY' and 'PRIMARY' in hdulist:
        data.meta[PRIHDR_KEY] = standardize_metadata(hdulist['PRIMARY'].header)
    data.meta.update(standardize_metadata(hdu.header))
    if include_wcs:
        data.coords = WCS(hdu.header, hdulist)
    component = Component.autotyped(hdu.data, units=bunit)
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=AstropyWarning)
        data.add_component(component=component, label=comp_label)

    return data


def _validate_bunit(bunit, raise_error=False):
    # TODO: Handle weird FITS BUNIT values here, as needed.
    if bunit == 'ELECTRONS/S':
        valid = 'electron/s'
    elif bunit == 'ELECTRONS':
        valid = 'electron'
    else:
        try:
            u.Unit(bunit)
        except Exception:
            if raise_error:
                raise
            valid = ''
        else:
            valid = bunit
    return valid


def _ndarray_to_glue_data(arr):
    if arr.ndim != 2:
        raise ValueError(f'Cannot load as image with ndim={arr.ndim}')

    data = Data()
    component = Component.autotyped(arr)
    data.add_component(component=component, label='DATA')
    return data


def _nddata_to_glue_data(ndd):
    if ndd.data.ndim != 2:
        raise ValueError(f'Cannot load as image with ndim={ndd.data.ndim}')

    returned_data = []
    for attrib in ('data', 'mask', 'uncertainty'):
        arr = getattr(ndd, attrib)
        if arr is None:
            returned_data.append(None)
            continue
        comp_label = attrib.upper()
        cur_data = Data()
        cur_data.meta.update(standardize_metadata(ndd.meta))
        if ndd.wcs is not None:
            cur_data.coords = ndd.wcs
        raw_arr = arr
        if attrib == 'data':
            bunit = ndd.unit or ''
        elif attrib == 'uncertainty':
            raw_arr = arr.array
            bunit = arr.unit or ''
        else:
            bunit = ''
        component = Component.autotyped(raw_arr, units=bunit)
        cur_data.add_component(component=component, label=comp_label)
        returned_data.append(cur_data)
    return returned_data


def _roman_asdf_2d_to_glue_data(file_obj, ext=None, try_gwcs_to_fits_sip=False):
    meta = standardize_roman_metadata(file_obj)
    gwcs = meta.get('wcs', None)

    if try_gwcs_to_fits_sip and gwcs is not None:
        coords = _try_gwcs_to_fits_sip(gwcs)
    else:
        coords = gwcs

    comp_label = ext.upper()
    data = Data(coords=coords)

    if HAS_ROMAN_DATAMODELS and isinstance(file_obj, (rdd.DataModel, rdd.ImageModel)):
        ext_values = file_obj[ext]
    else:
        ext_values = file_obj['roman'][ext]
    bunit = getattr(ext_values, 'unit', '')
    component = Component(np.array(ext_values), units=bunit)
    data.add_component(component=component, label=comp_label)
    data.meta.update(standardize_metadata(dict(meta)))

    return data


def _jwst2data(hdu, hdulist, try_gwcs_to_fits_sip=False):
    comp_label = hdu.name.lower()
    if comp_label.startswith("sci"):
        comp_label = 'data'
    data = Data()
    unit_attr = f'bunit_{comp_label}'

    try:
        # This is very specific to JWST pipeline image output.
        with asdf_in_fits.open(hdulist) as af:
            dm = af.tree
            dm_meta = af.tree["meta"]
            data.meta.update(standardize_metadata(dm_meta))

            if unit_attr in dm_meta:
                bunit = _validate_bunit(dm_meta[unit_attr], raise_error=False)
            else:
                bunit = ''

            # This is instance of gwcs.WCS, not astropy.wcs.WCS
            if 'wcs' in dm_meta:
                gwcs = dm_meta['wcs']

                if try_gwcs_to_fits_sip:
                    data.coords = _try_gwcs_to_fits_sip(gwcs)
                else:
                    data.coords = gwcs
            # keys in the asdf tree are lower case
            imdata = dm[comp_label]
            component = Component.autotyped(imdata, units=bunit)

            # Might have bad GWCS. If so, we exclude it.
            try:
                data.add_component(component=component, label=comp_label)

            except Exception:  # pragma: no cover
                data.coords = None
                data.add_component(component=component, label=comp_label)

    # TODO: Do not need this when jwst.datamodels finally its own package.
    # This might happen for grism image; fall back to FITS loader without WCS.
    except Exception:
        if comp_label == 'data':
            new_ext = 'sci'
        else:
            new_ext = hdu.name
        new_hdu = hdulist[new_ext]
        return _hdu2data(new_hdu, hdulist, include_wcs=False)

    return data



from jdaviz.core.helpers import ConfigHelper
from jdaviz.configs.default.plugins.line_lists.line_list_mixin import LineListMixin


class App(ConfigHelper, LineListMixin):
    """Base user-facing application helper."""

    _default_configuration = "deconfigged"  # temporary during deconfig process

    def __init__(self, *args, **kwargs):
        api_hints_obj = kwargs.pop('api_hints_obj', 'viz')
        super().__init__(*args, **kwargs)
        self.app.api_hints_obj = api_hints_obj

        # Temporary during deconfig process
        self.load = self._load
        self.app.state.dev_loaders = True



import regions
import numpy as np
from traitlets import Unicode, Bool, observe


from jdaviz.core.loaders.importers import BaseImporterToPlugin
from jdaviz.core.registries import loader_importer_registry
from jdaviz.core.template_mixin import AutoTextField
from jdaviz.core.user_api import ImporterUserApi


__all__ = ['FootprintImporter']


@loader_importer_registry('Footprint')
class FootprintImporter(BaseImporterToPlugin):
    template_file = __file__, "footprint.vue"

    footprint_label_value = Unicode().tag(sync=True)
    footprint_label_default = Unicode().tag(sync=True)
    footprint_label_auto = Bool(True).tag(sync=True)
    footprint_label_invalid_msg = Unicode().tag(sync=True)

    def __init__(self, app, resolver, parser, input, **kwargs):
        super().__init__(app, resolver, parser, input, **kwargs)
        self.footprint_label_default = 'default'
        self.footprint_label = AutoTextField(self, 'footprint_label_value',
                                             'footprint_label_default',
                                             'footprint_label_auto',
                                             'footprint_label_invalid_msg')

        self.observe(self._on_label_changed, 'footprint_label_value')
        self._on_label_changed()

    @property
    def is_valid(self):
        # TODO: handle str > region in parser

        def _ensure_sky(region):
            if isinstance(region, regions.Regions):
                return np.all([_ensure_sky(reg) for reg in region.regions])
            return hasattr(region, 'to_pixel')

        return (isinstance(self.input, (regions.Region, regions.Regions))
                and _ensure_sky(self.input)
                and self.has_default_plugin)

    @property
    def user_api(self):
        return ImporterUserApi(self, expose=['footprint_label'])

    @property
    def default_plugin(self):
        return 'Footprints'

    def _on_label_changed(self, msg={}):
        if not len(self.footprint_label_value.strip()):
            # strip will raise the same error for a label of all spaces
            self.footprint_label_invalid_msg = 'footprint_label must be provided'
            return

        if self.footprint_label_value == self.footprint_label_default:
            # _check_valid_footprint_label will say this is invalid,
            # once that changes, this block can be removed.
            self.footprint_label_invalid_msg = ''
            return

        self.footprint_label_invalid_msg = ''

    @observe('footprint_label_invalid_msg')
    def _set_import_disabled(self, change={}):
        self.import_disabled = len(self.footprint_label_invalid_msg) > 0

    def __call__(self):
        if self.footprint_label_invalid_msg:
            raise ValueError(self.footprint_label_invalid_msg)

        plg = self.app._jdaviz_helper.plugins['Footprints']
        if self.footprint_label_value not in plg.overlay.choices:
            # TODO: show warning in UI when entry already exists
            plg.add_overlay(self.footprint_label_value)
        plg.import_region(self.input)
